import numpy as np
import pyvista as pv
from scipy.spatial import cKDTree
from svv.domain.patch import Patch
from svv.domain.routines.allocate import allocate
from svv.domain.routines.discretize import contour
from svv.domain.io.read import read
from svv.domain.routines.tetrahedralize import tetrahedralize, triangulate
from svv.domain.routines.c_sample import pick_from_tetrahedron, pick_from_triangle, pick_from_line
from concurrent.futures import ProcessPoolExecutor, as_completed
from svv.domain.routines.boolean import boolean
# from svtoolkit.tree.utils.KDTreeManager import KDTreeManager
from svv.tree.utils.TreeManager import KDTreeManager, HNSWTree, USearchTree
from time import perf_counter
from tqdm import trange, tqdm
from sklearn.neighbors import BallTree
import random


class Domain(object):
    def __init__(self, *args, **kwargs):
        """
        The Domain class defines the region in space that
        will be recognized by svtoolkit when generating
        vascular networks. The class abstracts the physical
        representation of the space to allow for efficient
        interrogation and data manipulation.

        Parameters
        ----------
        args : list
            A list of arguments to be passed to the Domain
            object. The arguments can be a single numpy array
            of points, a single numpy array of points and
            normals, or a PyVista object.
        kwargs : dict
            A dictionary of keyword arguments to be passed
            to the Domain object.
        """
        self.patches = []
        self.functions = []
        self.function_tree = None
        self.points = None
        self.normals = None
        self.boundary = None
        self.original_boundary = None
        self.grid = None
        self.mesh = None
        self.random_seed = None
        self.random_generator = None
        self.characteristic_length = None
        self.mesh_tree = None
        self.boundary_nodes = None
        self.boundary_vertices = None
        self.mesh_nodes = None
        self.mesh_vertices = None
        self.convexity = None
        self.random_points = None
        if len(args) > 0:
            self.set_data(*args, **kwargs)

    def set_data(self, *args, **kwargs):
        """
        Set the data for the domain from point-wise data
        or a PyVista object.
        """
        if len(args) == 0:
            raise ValueError("No data provided.")
        elif len(args) == 1:
            if isinstance(args[0], np.ndarray):
                self.points = args[0]
                self.n = self.points.shape[0]
                self.d = self.points.shape[1]
            elif 'pyvista' in str(args[0].__class__):
                points, normals, n, d = read(args[0], **kwargs)
                self.original_boundary = args[0]
                self.points = points
                self.normals = normals
                self.n = n
                self.d = d
        elif len(args) == 2:
            self.points = args[0]
            self.normals = args[1]
            self.n = self.points.shape[0]
            self.d = self.points.shape[1]
        else:
            raise ValueError("Too many arguments.")
        return None

    def set_random_seed(self, seed):
        """
        Set the random seed for the domain.
        """
        self.random_seed = seed
        return None

    def set_random_generator(self):
        """
        Set the random generator for the domain.
        """
        self.random_generator = np.random.Generator(np.random.PCG64(seed=self.random_seed))
        return None

    def create(self, **kwargs):
        """
        Create the patches for the domain. This is the first step in the process.
        :param kwargs:
        :return:
        """
        self.patches = []
        if self.normals is None:
            patch_data = allocate(self.points, **kwargs)
        else:
            patch_data = allocate(self.points, self.normals, **kwargs)
        for i in trange(len(patch_data), desc='Creating patches', unit='patch', leave=False):
            self.patches.append(Patch())
            if self.normals is None:
                self.patches[-1].set_data(patch_data[i][0])
            else:
                self.patches[-1].set_data(patch_data[i][0], patch_data[i][1])
        return None

    def solve(self, **kwargs):
        """
        Solve the individual patch interpolation problems prior to blending.

        Parameters
        ----------
            None

        Returns
        -------
            None
        """
        for i in trange(len(self.patches), desc='Solving patches', unit='patch', leave=False):
            self.patches[i].solve(**kwargs)
        return None

    def build(self, **kwargs):
        """
        Build the implicit function describing the domain.
        :return:
        """
        resolution = kwargs.get('resolution', 25)
        skip_boundary = kwargs.get('skip_boundary', False)
        functions = []
        firsts = []
        for patch in self.patches:
            func = patch.build()
            firsts.append(func.first)
            functions.append(func)
        self.function_tree = cKDTree(np.array(firsts))
        function_list = set(tuple(list(range(len(functions)))))
        self.functions = functions
        # for fast evaluation
        a_shapes = np.array([function.a.shape for function in functions])
        b_shapes = np.array([function.b.shape for function in functions])
        c_shapes = np.array([function.c.shape for function in functions])
        d_values = np.array([function.d for function in functions]) # omit this since d is always a scalar
        pt_shapes = np.array([function.pts.shape for function in functions])
        self.A = np.full((len(functions), a_shapes[:, 0].max(), 1, 1, 1), np.nan)
        self.B = np.full((len(functions), b_shapes[:, 0].max(), 1, 1, 1, b_shapes[:, -1].max()), np.nan)
        self.C = np.full((len(functions), 1, 1, 1, c_shapes[:, -1].max()), np.nan)
        self.D = np.full((len(functions), 1, 1, 1), np.nan)
        self.PTS = np.full((len(functions), pt_shapes[:, 0].max(), 1, 1, 1, pt_shapes[:, -1].max()), np.nan)
        for i, function in enumerate(functions):
            self.A[i, :function.a.shape[0]] = function.a
            self.B[i, :function.b.shape[0]] = function.b
            self.C[i] = function.c
            self.D[i] = function.d
            self.PTS[i, :function.points.shape[0]] = function.pts
        # for fast evaluation
        if self.random_generator is None:
            self.set_random_generator()
        if not skip_boundary:
            self.get_boundary(resolution)
            self.get_interior()
        return None

    def __call__(self, points, **kwargs):
        """
        Evaluate the implicit function at a point or set of points.
        :param points:
        :param k:
        :param tolerance:
        :return:
        """
        return self.evaluate_fast(points, **kwargs)

    def __sub__(self, other):
        """
        Subtract two domains. This is the set difference operation.

        """
        new_domain = Domain()
        def evaluate(x):
            self_object = self.__call__(x)
            other_object = other.__call__(x)
            values = (np.logical_and(self_object < 0, other_object < 0)*(abs(self_object) + abs(other_object)) +
                      self_object)
            return values
        new_domain.evaluate = evaluate
        new_domain.points = np.vstack((self.points, other.points))
        new_domain.normals = np.vstack((self.normals, other.normals))
        #new_domain.evaluate = lambda x: np.logical_and(self.__call__(x) < 0, other.__call__(x) < 0)*(abs(self.__call__(x)) + abs(other.__call__(x))) + self.__call__(x)
        return new_domain

    def __add__(self, other):
        """
        Add two domains. This is the set union operation.
        """
        new_domain = Domain()
        def evaluate(x):
            self_object = self.__call__(x)
            other_object = other.__call__(x)
            values = (np.logical_and(self_object > 0, other_object > 0)*(abs(self_object) +
                                                                         abs(other_object)) +
                      (self_object < 0)*self_object + (other_object < 0)*other_object)
            return values
        new_domain.evaluate = evaluate
        new_domain.points = np.vstack((self.points, other.points))
        new_domain.normals = np.vstack((self.normals, other.normals))
        #new_domain.evaluate = lambda x: np.logical_and(self.__call__(x) > 0, other.__call__(x) > 0)*(abs(self.__call__(x)) + abs(other.__call__(x))) + (self.__call__(x) < 0)*self.__call__(x) + (other.__call__(x) < 0)*other.__call__(x)
        return new_domain

    def evaluate_fast(self, points, k=1, normalize=True, tolerance=np.finfo(float).eps * 4, show=False):
        if self.function_tree is None:
            raise ValueError("Domain not built.")
        if self.points.shape[1] != self.d:
            raise ValueError("Dimension mismatch.")
        values = np.zeros((points.shape[0], 1))
        dists, indices_first = self.function_tree.query(points, k=1)
        dists_first = dists.reshape(-1, 1)[:, -1].flatten()
        indices = self.function_tree.query_ball_point(points, dists_first + tolerance * dists_first)
        if normalize:
            normalize_scale = np.linalg.norm(np.max(self.points, axis=0) - np.min(self.points, axis=0))
        else:
            normalize_scale = 1
        if show:
            print("Points: ", points)
            print("Indices: ", indices)
            print("Distances: ", dists)
            print("First Indices: ", indices_first)
            print("First Distances: ", dists_first)
        indices_shape = np.array([len(indices[i]) for i in range(len(indices))])
        inds = np.full((len(indices), indices_shape.max()), -1)
        if show:
            print("Indices Shape: ", indices_shape)
            print("Inds Shape: ", inds.shape)
        for i in range(len(indices)):
            if len(indices[i]) == 0:
                inds[i, 0] = indices_first[i]
            else:
                inds[i, :len(indices[i])] = indices[i]
            #elif isinstance(indices_first[i], np.int64):
            #    print("Fallback indices found for point {} -> {}".format(i, points[i, :]))
            #    inds[i, 0] = indices_first[i]
            #else:
            #    print("No indices found for point {} -> {}".format(i, points[i, :]))
        inds_mask = np.ma.masked_array(inds, mask=(inds == -1))
        if np.any(np.all(inds_mask.mask, axis=1)):
            print("Mask for entire row! {}".format(np.argwhere(np.all(inds_mask.mask, axis=1))))
            print("Point: ", points[np.argwhere(np.all(inds_mask.mask, axis=1)).flatten(), :])
            print("Inds: ", inds[np.argwhere(np.all(inds_mask.mask, axis=1)).flatten() ,:])
        points_1 = points[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis, np.newaxis, :]
        points_2 = points[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis, :]
        diff = points_1 - self.PTS[inds.ravel()].reshape(inds.shape + tuple(list(self.PTS.shape)[1:]))
        diff_2 = np.nansum(np.square(diff), axis=-1)
        if show:
            print("Diff: ", diff)
            print("Diff^2: ", diff_2)
        # A Value
        value = np.nansum(self.A[inds.ravel()].reshape(inds.shape + tuple(list(self.A.shape)[1:])) * np.power(diff_2, (3 / 2)), axis=2)
        if show:
            print("A Value: ", value)
        # B Value
        value += np.nansum(3 * np.nansum(-(self.B[inds.ravel()].reshape(inds.shape + tuple(list(self.B.shape)[1:])))
                                         * diff, axis=-1) * np.power(diff_2, (3 / 2 - 1)), axis=2)
        if show:
            print("A + B Value: ", value)
        # C Value
        value += np.nansum(self.C[inds.ravel()].reshape(inds.shape + tuple(list(self.C.shape)[1:])) * points_2, axis=-1)
        if show:
            print("A + B + C Value: ", value)
        # D Value
        value += self.D[inds.ravel()].reshape(inds.shape + tuple(list(self.D.shape)[1:]))
        value = value.reshape(inds.shape)
        if show:
            print("A + B + C + D Value: ", value)
        if show:
            print("Values: ", value)
        if np.any(np.all(np.isnan(value), axis=1)):
            raise ValueError("Pre-mask All values are NaN in rows {}".format(np.argwhere(np.all(np.isnan(value), axis=1)).flatten()))
        value[inds_mask.mask] = np.nan
        if np.any(np.all(np.isnan(value), axis=1)):
            raise ValueError("All values are NaN in rows {}".format(np.argwhere(np.all(np.isnan(value), axis=1)).flatten()))
        sign = np.sign(np.nanmax(value, axis=1)).reshape(-1, 1)
        bools = np.isclose(value, 0)
        weights = np.ones(value.shape)
        weights[bools] = np.inf
        weights[~bools] = 1 / np.abs(value[~bools])
        weights = weights.reshape(value.shape)
        signs = np.sign(value)
        value[signs != sign] = np.nan
        values = np.zeros((points.shape[0], 1))
        values[sign.ravel() > 0, 0] = np.nanmax(value[sign.ravel() > 0, :], axis=1)
        values[sign.ravel() <= 0, 0] = np.nanmin(np.abs(value[sign.ravel() <= 0, :]), axis=1)
        values = np.tanh(sign.ravel() * np.sqrt(np.sum(np.square(values), axis=1)) / normalize_scale).reshape(-1, 1)
        return values

    def evaluate(self, points, k=1, normalize=True, tolerance=np.finfo(float).eps * 10, layer=None, show=False):
        """
        Evaluate the implicit function at a point or set of points.
        :param points:
        :param k:
        :param normalize
        :param tolerance:
        :return:
        """
        if self.function_tree is None:
            raise ValueError("Domain not built.")
        values = np.zeros((points.shape[0], 1))
        dists, indices_first = self.function_tree.query(points, k=1)
        dists_first = dists.reshape(-1, 1)[:, -1].flatten()
        indices = self.function_tree.query_ball_point(points, dists_first + tolerance*dists_first)
        if normalize:
            normalize_scale = np.linalg.norm(np.max(self.points, axis=0) - np.min(self.points, axis=0))
        else:
            normalize_scale = 1
        if show:
            print("Points: ", points)
            print("Indices: ", indices)
            print("Distances: ", dists)
            print("First Indices: ", indices_first)
            print("First Distances: ", dists_first)
        if k > 1:
            extra_dists, extra_indices_first = self.function_tree.query(points, k=k)
            extra_dists = extra_dists.reshape(-1, k)[:, -1].flatten()
            extra_indices = self.function_tree.query_ball_point(points, extra_dists + tolerance)
        else:
            extra_indices_first = None
            extra_indices = None
        for i in range(len(indices)):
            tmp_value = []
            weights = []
            if len(indices[i]) == 0:
                indices[i].append(indices_first[i])
            for j in range(len(indices[i])):
                if not show:
                    func = self.functions[indices[i][j]](points[i, :])
                else:
                    func = self.functions[indices[i][j]](points[i, :], show=True)
                tmp_value.append(func.flatten()[0])
            tmp_value = np.array(tmp_value)
            if show:
                print("Values: ", tmp_value)
            sign = np.sign(np.max(tmp_value))
            tmp_value_sign = np.argwhere(np.sign(tmp_value) == sign).flatten()
            #tmp_sum = np.sum(tmp_value[tmp_value_sign])
            bools = np.isclose(tmp_value, 0)
            weights = np.ones(tmp_value.shape[0])
            weights[bools] = np.inf
            weights[~bools] = 1 / np.abs(tmp_value[~bools])
            #if np.any(np.isclose(tmp_value, 0)):
            #    weights.append(np.inf)
            #else:
            #    weights.append(1 / np.min(abs(tmp_sum)))
            #tmp_value = [tmp_sum]
            #weights = [weights]
            #if k > 1:
            #    if len(extra_indices[i]) == 0:
            #        extra_indices[i].extend(extra_indices_first[i, :].tolist())
            #    for j in range(len(indices[i]), len(extra_indices[i])):
            #        func = self.functions[extra_indices[i][j]](points[i, :])
            #        if np.isclose(func.flatten()[0], 0):
            #            weights.append(np.inf)
            #            tmp_value.append(tmp_sum)
            #        else:
            #            weights.append(1 / abs(func.flatten()[0]))
            #            tmp_value.append(func.flatten()[0])
            #tmp_value = np.array(tmp_value)
            #weights = np.array(weights)
            signs = np.sign(tmp_value)
            idx = np.argwhere(signs == sign).flatten()
            if np.any(np.isinf(weights)):
                if sign > 0:
                    jdx = np.argmax(tmp_value[idx]).flatten()
                else:
                    jdx = np.argwhere(np.isinf(weights[idx])).flatten()
                values[i] = np.tanh(sign*np.sqrt(np.sum(np.square(tmp_value[idx][jdx])))/normalize_scale)
            else:
                if sign > 0:
                    jdx = np.argmax(tmp_value[idx]).flatten()
                else:
                    jdx = np.argmin(abs(tmp_value[idx])).flatten()
                #partition = weights[idx] / np.sum(weights[idx])
                value = np.tanh(sign*np.sqrt(np.sum(np.square(tmp_value[idx][jdx])))/normalize_scale)
                values[i] = value
        return values

    def within(self, points, level=0, **kwargs):
        """
        Determine if a point or set of points is within the domain.

        """
        level = kwargs.pop('layer', 0.0)
        values = self.__call__(points, **kwargs)
        return values <= level

    def get_boundary(self, resolution, **kwargs):
        """
        Descretize the domain into a set of points.
        """
        get_largest = kwargs.get('get_largest', True)
        if isinstance(self.original_boundary, type(None)):
            self.boundary, self.grid = contour(self.__call__, self.points, resolution)
        else:
            if not self.original_boundary.is_all_triangles:
                self.boundary = self.original_boundary.triangulate()
            else:
                self.boundary = self.original_boundary
            _, self.grid = contour(self.__call__, self.points, resolution)
        self.boundary = self.boundary.connectivity(extraction_mode='largest')
        self.boundary = self.boundary.compute_cell_sizes()
        if self.points.shape[1] == 2:
            self.boundary.cell_data['Normalized_Length'] = (self.boundary.cell_data['Length'] /
                                                            sum(self.boundary.cell_data['Length']))
            self.boundary_nodes = self.boundary.points.astype(np.float64)
            self.boundary_vertices = self.boundary.lines.reshape(-1, 3)[:, 1:].astype(np.int64)
        elif self.points.shape[1] == 3:
            self.boundary.cell_data['Normalized_Area'] = (self.boundary.cell_data['Area'] /
                                                          sum(self.boundary.cell_data['Area']))
            self.boundary_nodes = self.boundary.points.astype(np.float64)
            self.boundary_vertices = self.boundary.faces.reshape(-1, 4)[:, 1:].astype(np.int64)
        else:
            raise ValueError("Only 2D and 3D domains are supported.")
        return self.boundary, self.grid

    def get_interior(self, verbose=False, **kwargs):
        """
        Tetrahedralize the implicit function describing the domain

        Parameters
        ----------
        verbose : bool
            A flag to indicate if mesh fixing should be verbose.
        kwargs : dict
            A dictionary of keyword arguments to be passed to TetGen.

        Returns
        -------
        mesh : PyMesh mesh object
            The tetrahedralized mesh.
        """
        if self.boundary is None:
            raise ValueError("Boundary not defined. Call get_boundary() method first.")
        if self.points.shape[1] == 2:
            _mesh, nodes, vertices = triangulate(self.boundary, verbose=verbose, **kwargs)
            _mesh = _mesh.compute_cell_sizes()
            _mesh.cell_data['Normalized_Area'] = (_mesh.cell_data['Area'] / sum(_mesh.cell_data['Area']))
            self.all_mesh_cells = list(range(_mesh.n_cells))
            self.cumulative_probability = np.cumsum(_mesh.cell_data['Normalized_Area'])
            self.characteristic_length = _mesh.area**(1/self.points.shape[1])
            self.area = _mesh.area
            self.volume = 0.0
        elif self.points.shape[1] == 3:
            _mesh, nodes, vertices = tetrahedralize(self.boundary, verbose=verbose, **kwargs)
            _mesh = _mesh.compute_cell_sizes()
            _mesh.cell_data['Normalized_Volume'] = (_mesh.cell_data['Volume'] / sum(_mesh.cell_data['Volume']))
            self.all_mesh_cells = list(range(_mesh.n_cells))
            self.cumulative_probability = np.cumsum(_mesh.cell_data['Normalized_Volume'])
            self.characteristic_length = _mesh.volume**(1/self.points.shape[1])
            self.area = _mesh.area
            self.volume = _mesh.volume
        else:
            raise ValueError("Only 2D and 3D domains are supported.")
        self.mesh_tree = cKDTree(_mesh.cell_centers().points[:, :self.points.shape[1]], leafsize=4)
        self.mesh_tree_2 = BallTree(_mesh.cell_centers().points[:, :self.points.shape[1]])
        self.mesh = _mesh
        self.mesh_nodes = nodes.astype(np.float64)
        self.mesh_vertices = vertices.astype(np.int64)
        if self.points.shape[1] == 2:
            delaunay = pv.PolyData()
            tmp_points = np.zeros((self.points.shape[0], 3))
            tmp_points[:, :2] = self.points
            delaunay.points = tmp_points
            delaunay = delaunay.delaunay_2d(offset=2*np.linalg.norm(np.max(self.points, axis=0) -
                                                                    np.min(self.points, axis=0)))
            self.convexity = self.mesh.area / delaunay.area
        elif self.points.shape[1] == 3:
            delaunay = pv.PolyData()
            delaunay.points = np.unique(self.points, axis=0)
            delaunay = delaunay.delaunay_3d(offset=2*np.linalg.norm(np.max(self.points, axis=0) -
                                                                    np.min(self.points, axis=0)))
            self.convexity = self.mesh.volume / delaunay.volume
        else:
            raise ValueError("Only 2D and 3D domains are supported.")
        return _mesh

    def get_interior_points(self, n, tree=None, volume_threshold=None,
                            threshold=None, method=None, implicit_range=(-1.0, 0.0), **kwargs):
        """
        Pick n points randomly from the domain.
        """
        use_random_int = kwargs.get('use_random_int', False)
        convex = kwargs.get('convex', False)
        if self.mesh is None or method == 'implicit_only':
            min_dims = np.min(self.points, axis=0)
            max_dims = np.max(self.points, axis=0)
            points = np.ones((n, self.points.shape[1]), dtype=np.float64)*np.nan
            remaining_points = n
            while remaining_points > 0:
                tmp_points = ((self.random_generator.random((n, self.points.shape[1]))-0.5) *
                              (max_dims - min_dims).reshape(1, -1) + (max_dims + min_dims).reshape(1, -1)/2)
                values = self.__call__(tmp_points[:, :self.points.shape[1]]).flatten()
                tmp_points = tmp_points[values < implicit_range[1], :]
                values = values[values < implicit_range[1]]
                tmp_points = tmp_points[values > implicit_range[0], :]
                added_points = min(remaining_points, tmp_points.shape[0])
                points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points,
                                                                                                 :self.points.shape[1]]
                remaining_points -= added_points
            cells = np.ones((n,), dtype=np.int64)*-1
        elif method == 'preallocate':
            # random tree structure for selecting points
            if isinstance(self.random_points,type(None)) or self.random_points.shape[0] < 2*n:
                pts, _ = self.get_interior_points(10*n)
                self.random_points = pts
            points = np.ones((n, self.points.shape[1]), dtype=np.float64) * np.nan
            remaining_points = n
            while remaining_points > 0:
                pt_dists, pt_ids = tree.query(self.random_points)
                if not isinstance(threshold, type(None)) and not isinstance(volume_threshold, type(None)):
                    mask = np.logical_and(pt_dists > threshold, pt_dists < volume_threshold)
                else:
                    mask = pt_dists > 0.0
                tmp_points = self.random_points[mask.flatten(),:]
                added_points = min(remaining_points, tmp_points.shape[0])
                points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points, :]
                remaining_points -= added_points
                if remaining_points > 0:
                    pts, _ = self.get_interior_points(10 * n)
                    self.random_points = pts
            cells = np.ones((n,), dtype=np.int64) * -1
        else:
            replace = kwargs.get('replace', True)
            points = np.ones((n, self.points.shape[1]), dtype=np.float64) * np.nan
            remaining_points = n
            ball_point = 0
            set_calc = 0
            choice_calc = 0
            domain_calc = 0
            while remaining_points > 0:
                if self.points.shape[1] == 3:
                    #if isinstance(tree, KDTreeManager) and isinstance(threshold, float) and not convex:
                    if isinstance(threshold, float) and not convex:
                        #cells_outer = []
                        start = perf_counter()
                        #cells_0 = tree.query_ball_tree(self.mesh_tree, volume_threshold, eps=volume_threshold/100)
                        #start = perf_counter()
                        if volume_threshold is None:
                            cells_outer = np.arange(self.mesh.n_cells, dtype=np.int64)
                        else:
                            #cells_0 = self.mesh_tree_2.query_radius(tree.active_tree.data, volume_threshold)
                            cells_0 = self.mesh_tree_2.query_radius(tree, volume_threshold)
                            cells_outer = np.unique(np.concatenate(cells_0))
                        #_ = [cells_outer.extend(cell) for cell in cells_0]
                        #cells_1 = tree.query_ball_tree(self.mesh_tree, threshold, eps=threshold/100)
                        #cells_1 = self.mesh_tree_2.query_radius(tree.active_tree.data, threshold)
                        cells_1 = self.mesh_tree_2.query_radius(tree, threshold)
                        #cells_inner = []
                        #_ = [cells_inner.extend(cell) for cell in cells_1]
                        cells_inner = np.unique(np.concatenate(cells_1))
                        #end = perf_counter()
                        #ball_point += end - start
                        #start = perf_counter()
                        #cells = np.array(list(cells_outer - cells_inner))
                        #_, idx = self.mesh_tree.query_ball_point(tree.active_tree.data, k=min(100, self.mesh.n_cells))
                        #cells = np.unique(idx[:, 50:].flatten())
                        cells = np.setdiff1d(cells_outer, cells_inner)
                        end = perf_counter()
                        set_calc += end - start
                        start = perf_counter()
                        if len(cells) == 0:
                            if not use_random_int:
                                #cells = self.random_generator.choice(list(range(self.mesh.n_cells)), n,
                                #                                     p=self.mesh.cell_data['probability'],
                                #                                     replace=replace)
                                cells = np.array(random.choices(self.all_mesh_cells,
                                                       cum_weights=self.cumulative_probability,k=n))
                            else:
                                cells = self.random_generator.integers(0, self.mesh.n_cells, n)
                        else:
                            if not use_random_int:
                                #cells = self.random_generator.choice(cells, n,
                                #                                     p=(self.mesh.cell_data['probability'][cells] /
                                #                                        np.sum(self.mesh.cell_data['probability'][cells])),
                                #                                     replace=replace)
                                cumulative_probability = np.cumsum(self.mesh.cell_data['Normalized_Volume'][cells])
                                cells = np.array(random.choices(cells.tolist(),
                                                                cum_weights=cumulative_probability, k=n))
                            else:
                                cells = self.random_generator.choice(cells, n, replace=True)
                        end = perf_counter()
                        choice_calc += end - start
                    else:
                        start = perf_counter()
                        if not use_random_int:
                            #cells = self.random_generator.choice(list(range(self.mesh.n_cells)), n,
                            #                                     p=self.mesh.cell_data['probability'],
                            #                                     replace=replace)
                            cells = np.array(random.choices(self.all_mesh_cells,
                                                            cum_weights=self.cumulative_probability, k=n))
                        else:
                            cells = self.random_generator.integers(0, self.mesh.n_cells, n)
                        end = perf_counter()
                        choice_calc += end - start
                        #if use_random_int:
                        #    print("Time from random int: ", end - start)
                    start = perf_counter()
                    rdx = self.random_generator.random((n, 4, 1))
                    simplices = self.mesh_nodes[self.mesh_vertices[cells, :], :]
                    tmp_points = pick_from_tetrahedron(simplices, rdx)
                    if implicit_range[1] == 0 and implicit_range[0] == -1:
                        pass
                    else:
                        values = self.__call__(tmp_points).flatten()
                        tmp_points = tmp_points[values <= implicit_range[1], :]
                        values = values[values <= implicit_range[1]]
                        tmp_points = tmp_points[values >= implicit_range[0], :]
                    added_points = min(remaining_points, tmp_points.shape[0])
                    points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points, :]
                    remaining_points -= added_points
                    end = perf_counter()
                    domain_calc += end - start
                elif self.points.shape[1] == 2:
                    cells = self.random_generator.choice(list(range(self.mesh.n_cells)), n,
                                                         p=self.mesh.cell_data['Normalized_Area'],
                                                         replace=replace)
                    rdx = self.random_generator.random((n, 3, 1))
                    simplices = self.mesh_nodes[self.mesh_vertices[cells, :], :]
                    tmp_points = pick_from_triangle(simplices, rdx)
                    values = self.__call__(tmp_points[:, :2]).flatten()
                    tmp_points = tmp_points[values <= implicit_range[1], :]
                    tmp_points = tmp_points[values >= implicit_range[0], :]
                    added_points = min(remaining_points, tmp_points.shape[0])
                    points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points, :2]
                    remaining_points -= added_points
            #if tree is not None and tree.active_tree.data.shape[0] <= 3 and not convex:
            #    mesh_cells = np.setdiff1d(cells_outer, cells_inner)
            #    if mesh_cells.shape[0] > 0:
            #        plotter = pv.Plotter()
            #        plotter.add_mesh(self.mesh, color='white', opacity=0.25)
            #        plotter.add_mesh(self.mesh.extract_cells(mesh_cells), color='red', opacity=0.5)
            #        plotter.add_points(points, color='blue', point_size=5)
            #        if isinstance(tree.active_tree, cKDTree) and isinstance(threshold, float):
            #            plotter.add_points(tree.active_tree.data, color='green', point_size=10)
            #        plotter.show()
            #if ball_point > 0.01:
            #    print(f'Ball Point took {ball_point} seconds')
            #if set_calc > 0.01:
            #    print(f'Set Calculation took {set_calc} seconds')
            #if choice_calc > 0.01:
            #    print(f'Choice Calculation took {choice_calc} seconds')
            #if domain_calc > 0.01:
            #    print(f'Domain Calculation took {domain_calc} seconds')
        return points, cells

    def get_boundary_points(self, n, method=None, **kwargs):
        """
        Pick n points randomly from the boundary of the domain.
        """
        if self.mesh is None or method == 'implicit_only':
            raise NotImplementedError("Implicit only not implemented.")
        else:
            replace = kwargs.get('replace', True)
            points = np.ones((n, self.points.shape[1]), dtype=np.float64) * np.nan
            remaining_points = n
            while remaining_points > 0:
                if self.points.shape[1] == 3:
                    cells = self.random_generator.choice(list(range(self.boundary.n_cells)), n,
                                                         p=self.boundary.cell_data['Normalized_Area'],
                                                         replace=replace)
                    rdx = self.random_generator.random((n, 3, 1))
                    simplices = self.boundary_nodes[self.boundary_vertices[cells, :], :]
                    tmp_points = pick_from_triangle(simplices, rdx)
                    added_points = min(remaining_points, tmp_points.shape[0])
                    points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points, :]
                    remaining_points -= added_points
                elif self.points.shape[1] == 2:
                    cells = self.random_generator.choice(list(range(self.boundary.n_cells)), n,
                                                         p=self.boundary.cell_data['Normalized_Length'],
                                                         replace=replace)
                    rdx = self.random_generator.random((n, 2, 1))
                    simplices = self.boundary_nodes[self.boundary_vertices[cells, :], :]
                    tmp_points = pick_from_line(simplices, rdx)
                    added_points = min(remaining_points, tmp_points.shape[0])
                    points[n - remaining_points:n - remaining_points + added_points, :] = tmp_points[:added_points, :2]
                    remaining_points -= added_points
        return points

    def get_exterior_points(self, n, method=None, implicit_range=(0.0, 1.0), **kwargs):
        raise NotImplementedError("Not implemented.")



make_results = """import pyvista as pv
import numpy as np
import os
from tqdm import tqdm
import csv
from collections import defaultdict, OrderedDict


def read_csv_into_nested_dict(file_path):
    # Initialize a nested dictionary structure
    data = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    time_steps = defaultdict(lambda: defaultdict(dict))
    time_counter = defaultdict(int)
    with open(file_path, mode='r', newline='', encoding='utf-8') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            name = row['name']
            time = float(row['time'])
            flow_in = float(row['flow_in'])
            flow_out = float(row['flow_out'])
            pressure_in = float(row['pressure_in'])
            pressure_out = float(row['pressure_out'])
            parts = name.split('_')
            branch = int(parts[0].replace('branch', ''))
            segment = int(parts[1].replace('seg', ''))
            # Check if the time has already been encountered for this branch and segment
            if time not in time_steps[branch][segment]:
                time_steps[branch][segment][time] = time_counter[(branch, segment)]
                time_counter[(branch, segment)] += 1
            time_index = time_steps[branch][segment][time]
            data['time'][branch][segment][time_index] = time
            data['flow_in'][branch][segment][time_index] = flow_in
            data['flow_out'][branch][segment][time_index] = flow_out
            data['pressure_in'][branch][segment][time_index] = pressure_in
            data['pressure_out'][branch][segment][time_index] = pressure_out
    return data


geom_data = np.genfromtxt("geom.csv",delimiter=",")
data = read_csv_into_nested_dict("output.csv")

total = None
timepoints = []
min_pressure = np.inf
max_pressure = -np.inf
min_flow = np.inf
max_flow = -np.inf
min_wss = np.inf
max_wss = -np.inf
for idx in tqdm(range(len(data['time'][0][0])),desc="Building Timeseries ",position=0):
    #time_merge = None
    time = data['time'][0][0][idx]
    tmp_vessels = []
    for jdx in tqdm(range(len(data['flow_in'])),desc="Building Vessel Data",position=1,leave=False):
        vessel = list(data['flow_in'].keys())[jdx]
        start = geom_data[vessel,0:3]
        end   = geom_data[vessel,3:6]
        direction = (geom_data[vessel,3:6] - geom_data[vessel,0:3])/np.linalg.norm(geom_data[vessel,3:6] - geom_data[vessel,0:3])
        length    = geom_data[vessel,6]
        radius    = geom_data[vessel,7]
        number_segments = len(data['flow_in'][vessel]) #- 1 #assume only 1 segment right now
        number_points   = len(data['flow_in'][vessel])
        vdx = vessel
        for kdx in range(number_segments):
            center = (1/2)*direction*length + start
            vessel = pv.Cylinder(center=center,direction=direction,height=length,radius=radius)
            vessel = vessel.elevation(low_point=end,high_point=start,scalar_range=[data['pressure_out'][vdx][kdx][idx]/1333.33, data['pressure_in'][vdx][kdx][idx]/1333.33])
            if data['pressure_in'][vdx][kdx][idx]/1333.33 > max_pressure:
                max_pressure = data['pressure_in'][vdx][kdx][idx]/1333.33
            if data['pressure_out'][vdx][kdx][idx]/1333.33 < min_pressure:
                min_pressure = data['pressure_out'][vdx][kdx][idx]/1333.33
            vessel.rename_array('Elevation','Pressure [mmHg]',preference='point')
            vessel.cell_data['Flow [mL/s]'] = data['flow_in'][vdx][kdx][idx]
            re = (1.06*2*radius*((data['flow_in'][vdx][kdx][idx]/(np.pi*radius**2))/2))/0.04
            fd = 64/re
            wss = ((data['flow_in'][vdx][kdx][idx]/(np.pi*radius**2))/2)*fd*1.06
            vessel.cell_data['WSS [dyne/cm^2]']  = wss
            if max_flow < data['flow_in'][vdx][kdx][idx]:
                max_flow = data['flow_in'][vdx][kdx][idx]
            if min_flow > data['flow_in'][vdx][kdx][idx]:
                min_flow = data['flow_in'][vdx][kdx][idx]
            if max_wss < wss:
                max_wss = wss
            if min_wss > wss:
                min_wss = wss
            tmp_vessels.append(vessel)
        #if time_merge is None:
        #    time_merge = vessel
        #else:
        #    time_merge = time_merge.merge(vessel)
    time_merge = tmp_vessels[0].merge(tmp_vessels[1:])
    time_merge.field_data['time'] = time
    timepoints.append(time_merge)
    #if total is None:
    #    total = time_merge
    #else:
    #    total = total.merge(time_merge)
    if not os.path.isdir("timeseries"):
        os.mkdir("timeseries")

if not os.path.isdir("timeseries_for_pressure_gif"):
    os.mkdir("timeseries_for_pressure_gif")
if not os.path.isdir("timeseries_for_flow_gif"):
    os.mkdir("timeseries_for_flow_gif")
if not os.path.isdir("timeseries_for_wss_gif"):
    os.mkdir("timeseries_for_wss_gif")
total = timepoints[0].merge(timepoints[1:])
for i in tqdm(range(len(timepoints)),desc="Saving Timeseries",position=1):
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='Pressure [mmHg]',clim=[round(min_pressure,4),round(max_pressure,4)],cmap="coolwarm")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_pressure_gif"+os.sep+"time_point_{}.png".format(i))
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='Flow [mL/s]',clim=[round(min_flow,4),round(max_flow,4)],cmap="GnBu")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_flow_gif"+os.sep+"time_point_{}.png".format(i))
    p = pv.Plotter(off_screen=True)
    p.add_mesh(timepoints[i],scalars='WSS [dyne/cm^2]',clim=[round(min_wss,2),round(max_wss,2)],cmap="coolwarm")
    p.show(auto_close=True,screenshot=os.getcwd()+os.sep+"timeseries_for_wss_gif"+os.sep+"time_point_{}.png".format(i))
    timepoints[i].save(os.getcwd()+os.sep+"timeseries"+os.sep+"time_point_{}.vtp".format(i))

total.save(os.getcwd()+os.sep+"timeseries"+os.sep+"total.vtp")

"""

view_plots="""import numpy as np
import matplotlib.pyplot as plt
data = np.load("solver_0d_branch_results.npy",allow_pickle=True).item()
vessels = []
fig_flow = plt.figure()
ax_flow = fig_flow.add_subplot()
fig_pressure = plt.figure()
ax_pressure = fig_pressure.add_subplot()
fig_flow_outlets = plt.figure()
ax_flow_outlets = fig_flow_outlets.add_subplot()
fig_pressure_outlets = plt.figure()
ax_pressure_outlets = fig_pressure_outlets.add_subplot()
for vessel in data["flow"]:
    vessels.append(vessel)
for vessel in vessels:
    ax_flow.plot(data["time"],data["flow"][vessel][0],label="vessel_"+str(vessel))
    ax_pressure.plot(data["time"],data["pressure"][vessel][0]/1333.22,label="vessel_"+str(vessel))
import json
info = json.load(open("solver_0d.in"))
all_inlets = []
all_outlets = []
for i in info["junctions"]:
    for j in i["inlet_vessels"]:
        all_inlets.append(j)
    for j in i["outlet_vessels"]:
        all_outlets.append(j)
all_inlets = set(all_inlets)
all_outlets = set(all_outlets)
true_outlets = list(all_outlets.difference(all_inlets))
true_inlets = list(all_inlets.difference(all_outlets))
for vessel in true_outlets:
    ax_flow_outlets.plot(data["time"],data["flow"][vessel][-1],label="vessel_"+str(vessel))
    ax_pressure_outlets.plot(data["time"],data["pressure"][vessel][-1]/1333.22,label="vessel_"+str(vessel))
ax_flow.set_xlabel("Time (sec)")
ax_flow.set_ylabel("Flow (mL/s)")
ax_pressure.set_xlabel("Time (sec)")
ax_pressure.set_ylabel("Pressure (mmHg)")
ax_flow_outlets.set_xlabel("Time (sec)")
ax_pressure_outlets.set_xlabel("Time (sec)")
ax_flow_outlets.set_ylabel("Flow (mL/s)")
ax_flow_outlets.set_title("Outlets Only")
ax_pressure_outlets.set_ylabel("Pressure (mmHg)")
ax_pressure_outlets.set_title("Outlets Only")
#ax_pressure_outlets.set_ylim([-10*np.finfo(float).eps,10*np.finfo(float).eps])
plt.show()

"""

post_data = """import pyvista as pv
import numpy as np
import os
from tqdm import tqdm
import csv
from collections import defaultdict, OrderedDict


def read_csv_into_nested_dict(file_path):
    # Initialize a nested dictionary structure
    data = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    time_steps = defaultdict(lambda: defaultdict(dict))
    time_counter = defaultdict(int)
    with open(file_path, mode='r', newline='', encoding='utf-8') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            name = row['name']
            time = float(row['time'])
            flow_in = float(row['flow_in'])
            flow_out = float(row['flow_out'])
            pressure_in = float(row['pressure_in'])
            pressure_out = float(row['pressure_out'])
            parts = name.split('_')
            branch = int(parts[0].replace('branch', ''))
            segment = int(parts[1].replace('seg', ''))
            # Check if the time has already been encountered for this branch and segment
            if time not in time_steps[branch][segment]:
                time_steps[branch][segment][time] = time_counter[(branch, segment)]
                time_counter[(branch, segment)] += 1
            time_index = time_steps[branch][segment][time]
            data['time'][branch][segment][time_index] = time
            data['flow_in'][branch][segment][time_index] = flow_in
            data['flow_out'][branch][segment][time_index] = flow_out
            data['pressure_in'][branch][segment][time_index] = pressure_in
            data['pressure_out'][branch][segment][time_index] = pressure_out
    return data


geom_data = np.genfromtxt("geom.csv",delimiter=",")
data = read_csv_into_nested_dict("output.csv")

# Storing 5 data results: time, pressure_in, pressure_out, flow_in, flow_out
results = np.zeros((geom_data.shape[0], geom_data.shape[1] + 5, len(data['time'][0][0])))
for idx in tqdm(range(len(data['time'][0][0])),desc="Building Timeseries ",position=0):
    time = data['time'][0][0][idx]
    # Assign data from the csv file to the vessels in the results array for a given time point
    for jdx in range(results.shape[0]):
        results[jdx, 0:3, idx] = geom_data[jdx, 0:3]
        results[jdx, 3:6, idx] = geom_data[jdx, 3:6]
        results[jdx, 6, idx] = geom_data[jdx, 6]
        results[jdx, 7, idx] = geom_data[jdx, 7]
        results[jdx, 8, idx] = data['time'][jdx][0][idx]
        results[jdx, 9, idx] = data['pressure_in'][jdx][0][idx]
        results[jdx, 10, idx] = data['pressure_out'][jdx][0][idx]
        results[jdx, 11, idx] = data['flow_in'][jdx][0][idx]
        results[jdx, 12, idx] = data['flow_out'][jdx][0][idx]

# Save the results array to a numpy file
np.save("results.npy", results)
"""


from abc import ABC

import numpy as np
from svv.forest.connect.base import BaseCurve


class BezierCurve(BaseCurve, ABC):
    """
    A concrete implementation of Curve for Bézier curves in 2D or 3D space.

    Attributes
    ----------
    control_points : np.ndarray
        Array of shape (n+1, d), where n is the polynomial degree of the Bézier curve,
        and d = 2 or 3 for 2D or 3D curves.
    """

    def __init__(self, control_points):
        """
        Initialize the BezierCurve with control points.

        Parameters
        ----------
        control_points : array-like
            A 2D or 3D array/list of shape (n+1, d), where n+1 is the number
            of control points, and d is either 2 or 3.
        """
        self.control_points = np.array(control_points, dtype=float)
        if self.control_points.ndim != 2:
            raise ValueError("control_points must be a 2D array of shape (n+1, d).")

    def _de_casteljau(self, points, t):
        """
        Compute a single point on the Bézier curve using De Casteljau's algorithm.

        Parameters
        ----------
        points : np.ndarray
            Control points (or derivative control points) to be interpolated,
            shape (m, d).
        t : float
            Parametric value in the range [0, 1].

        Returns
        -------
        np.ndarray
            A single point on the Bézier curve at the given parametric value,
            shape (d,).
        """
        pts = points.copy()
        while len(pts) > 1:
            pts = (1 - t) * pts[:-1] + t * pts[1:]
        return pts[0]

    def evaluate(self, t_values):
        """
        Evaluate the Bézier curve at the given array of parametric values.

        Parameters
        ----------
        t_values : array-like
            Parametric values in the range [0, 1].

        Returns
        -------
        np.ndarray
            Points on the Bézier curve at the specified parametric values.
            Shape: (len(t_values), d).
        """
        t_values = np.atleast_1d(t_values)
        return np.array([self._de_casteljau(self.control_points, t) for t in t_values])

    def derivative(self, t_values, order=1):
        """
        Compute the nth derivative of the Bézier curve at specified parametric values.

        Parameters
        ----------
        t_values : array-like
            Parametric values in the range [0, 1].
        order : int, optional
            The order of the derivative (1 for first derivative, 2 for second, etc.).

        Returns
        -------
        np.ndarray
            The derivative values at each parametric value, shape (len(t_values), d).
        """
        n = len(self.control_points) - 1
        derivative_points = self.control_points.copy()

        # Iteratively compute up to the requested derivative order
        for _ in range(order):
            derivative_points = n * (derivative_points[1:] - derivative_points[:-1])
            n -= 1
            if n < 0:
                # Instead of np.zeros_like(...), create a shape (1, d) array of zeros
                derivative_points = np.zeros((1, derivative_points.shape[1]))
                break

        t_values = np.atleast_1d(t_values)
        return np.array([self._de_casteljau(derivative_points, t) for t in t_values])

    def roc(self, t_values):
        """
        Compute the radius of curvature of the Bézier curve at specified parametric values.

        Parameters
        ----------
        t_values : array-like
            Parametric values in the range [0, 1].

        Returns
        -------
        np.ndarray
            Radius of curvature values along the Bézier curve.
            Shape: (len(t_values),).

        Notes
        -----
        Formula: R = |v|^3 / |v x a|, where
            - v is the first derivative (velocity),
            - a is the second derivative (acceleration).
        """
        t_values = np.atleast_1d(t_values)
        v = self.derivative(t_values, order=1)  # shape (N, d)
        a = self.derivative(t_values, order=2)  # shape (N, d)

        v_mags = np.linalg.norm(v, axis=1)

        if self.dimension == 2:
            # Cross product in 2D is a scalar: v_x a_y - v_y a_x
            cross_va = v[:, 0] * a[:, 1] - v[:, 1] * a[:, 0]
            cross_mags = np.abs(cross_va)
        else:
            # 3D cross product => shape (N, 3)
            cross_va = np.cross(v, a)
            cross_mags = np.linalg.norm(cross_va, axis=1)

        epsilon = 1e-12
        roc_values = v_mags ** 3 / (cross_mags + epsilon)
        return roc_values

    def torsion(self, t_values):
        """
        Compute the torsion of the Bézier curve at specified parametric values.

        Parameters
        ----------
        t_values : array-like
            Parametric values in the range [0, 1].

        Returns
        -------
        np.ndarray
            Torsion values at each parametric value.
            Shape: (len(t_values),).

        Notes
        -----
        Torsion is generally meaningful in 3D. In 2D, it's typically zero.
        Formula: τ = ((v x a) · b) / |v x a|^2, where
            - v: first derivative
            - a: second derivative
            - b: third derivative
        """
        t_values = np.atleast_1d(t_values)

        # If 2D, short-circuit to zero
        if self.dimension == 2:
            return np.zeros_like(t_values, dtype=float)

        # 3D case
        v = self.derivative(t_values, order=1)
        a = self.derivative(t_values, order=2)
        b = self.derivative(t_values, order=3)

        cross_va = np.cross(v, a)  # shape (N,3)
        numerator = np.einsum('ij,ij->i', cross_va, b)  # dot((v x a), b)
        denominator = np.einsum('ij,ij->i', cross_va, cross_va)  # |v x a|^2

        with np.errstate(divide='ignore', invalid='ignore'):
            torsion_values = numerator / denominator
            torsion_values = np.nan_to_num(torsion_values)

        return torsion_values

    def arc_length(self, t_start=0, t_end=1, num_points=100):
        """
        Compute the approximate arc length of the Bézier curve between two parametric values.

        Parameters
        ----------
        t_start : float, optional
            Start parameter in [0, 1]. Default is 0.
        t_end : float, optional
            End parameter in [0, 1]. Default is 1.
        num_points : int, optional
            Number of sample points for piecewise linear approximation. Default is 100.

        Returns
        -------
        float
            Approximate arc length over [t_start, t_end].
        """
        if not (0 <= t_start < t_end <= 1):
            raise ValueError("t_start and t_end must satisfy 0 <= t_start < t_end <= 1.")

        # Sample parametric values
        t_values = np.linspace(t_start, t_end, num_points)

        # Evaluate points on the curve
        points = self.evaluate(t_values)

        # Sum distances between consecutive sample points
        segment_lengths = np.linalg.norm(np.diff(points, axis=0), axis=1)
        arc_len = np.sum(segment_lengths)

        return arc_len

    @property
    def dimension(self):
        return self.control_points.shape[1]



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython

cnp.import_array()


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def pick_from_tetrahedron(double[:,:,:] simplicies, double[:,:,:] rdx):
    cdef Py_ssize_t i = simplicies.shape[0]
    cdef cnp.ndarray[double, ndim=2] points = np.zeros((i, 3), dtype=np.float64)
    cdef double tmp
    for ii in range(i):
        if rdx[ii, 1, 0] +  rdx[ii, 2, 0] > 1:
            rdx[ii, 1, 0] = 1.0 - rdx[ii, 1, 0]
            rdx[ii, 2, 0] = 1.0 - rdx[ii, 2, 0]
        if rdx[ii, 2, 0] + rdx[ii, 3, 0] > 1:
            tmp = rdx[ii, 3, 0]
            rdx[ii, 3, 0] = 1.0 - rdx[ii, 1, 0] - rdx[ii, 2, 0]
            rdx[ii, 2, 0] = 1.0 - tmp
        elif rdx[ii, 1, 0] + rdx[ii, 2, 0] + rdx[ii, 3, 0] > 1:
            tmp = rdx[ii, 3, 0]
            rdx[ii, 3, 0] = rdx[ii, 1, 0] + rdx[ii, 2, 0] + rdx[ii, 3, 0] - 1.0
            rdx[ii, 1, 0] = 1.0 - rdx[ii, 2, 0] - tmp
        rdx[ii, 0, 0] = 1.0 - rdx[ii, 1, 0] - rdx[ii, 2, 0] - rdx[ii, 3, 0]
    points = np.sum(np.multiply(rdx, simplicies), axis=1)
    return points


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def pick_from_triangle(double[:,:,:] simplices, double[:,:,:] rdx):
    cdef Py_ssize_t i = simplices.shape[0]
    cdef cnp.ndarray[double, ndim=2] points = np.zeros((i, 3), dtype=np.float64)
    for ii in range(i):
        if rdx[ii, 1, 0] +  rdx[ii, 2, 0] > 1:
            rdx[ii, 1, 0] = 1.0 - rdx[ii, 1, 0]
            rdx[ii, 2, 0] = 1.0 - rdx[ii, 2, 0]
        rdx[ii, 0, 0] = 1.0 - rdx[ii, 1, 0] - rdx[ii, 2, 0]
    points = np.sum(np.multiply(rdx, simplices), axis=1)
    return points

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def pick_from_line(double[:,:,:] simplices, double[:,:,:] rdx):
    cdef Py_ssize_t i = simplices.shape[0]
    cdef cnp.ndarray[double, ndim=2] points = np.zeros((i, 3), dtype=np.float64)
    for ii in range(i):
        rdx[ii, 0, 0] = 1.0 - rdx[ii, 1, 0]
    points = np.sum(np.multiply(rdx, simplices), axis=1)
    return points


from xml.dom import minidom

from svv.simulation.output import Output
from svv.simulation.mesh import GeneralMesh
from svv.simulation.boundary_conditions import BoundaryCondition
from svv.simulation.linear_solver import LinearSolver


class EquationBase:
    def __init__(self):
        self.equation_type = None
        self.coupled = False
        self.min_iterations = None
        self.max_iterations = None
        self.tolerance = None
        self.backflow_stabilization_coefficient = None
        self.density = None
        self.conductivity = None
        self.source_term = None
        self.viscosity = None
        self.linear_solver = LinearSolver()
        self.outputs = []
        self.alias = None

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return str(vars(self))

    def __eq__(self, other):
        check = ["equation_type", "coupled", "min_iterations", "max_iterations", "tolerance",
                 "backflow_stabilization_coefficient", "density", "conductivity", "source_term"]
        attributes = vars(self)
        other_attributes = vars(other)
        return all(other_attributes[key] == attributes[key] for key in check)

    def __ne__(self, other):
        return not self.__eq__(other)


class Equation(EquationBase):
    def __init__(self):
        super().__init__()
        self.file = minidom.Document()
        self.meshes = []
        self.boundary_conditions = []
        self.faces = []

    def __str__(self):
        return self.toxml().toprettyxml()

    def __repr__(self):
        return self.toxml().toprettyxml()

    def set_type(self, equation_type):
        if equation_type not in ["fluid", "scalarTransport"]:
            raise ValueError("Equation type must be 'fluid' or 'scalarTransport'.")
        self.equation_type = equation_type

    def set_coupled(self, coupled):
        if not isinstance(coupled, bool):
            raise ValueError("Coupled must be a boolean.")
        self.coupled = coupled

    def set_min_iterations(self, min_iterations):
        if not isinstance(min_iterations, int):
            raise ValueError("Min iterations must be an integer.")
        self.min_iterations = min_iterations

    def set_max_iterations(self, max_iterations):
        if not isinstance(max_iterations, int):
            raise ValueError("Max iterations must be an integer.")
        self.max_iterations = max_iterations

    def set_tolerance(self, tolerance):
        if not isinstance(tolerance, float):
            raise ValueError("Tolerance must be a float.")
        self.tolerance = tolerance

    def set_backflow_stabilization_coefficient(self, backflow_stabilization_coefficient):
        if not isinstance(backflow_stabilization_coefficient, float):
            raise ValueError("Backflow stabilization coefficient must be a float.")
        self.backflow_stabilization_coefficient = backflow_stabilization_coefficient

    def set_density(self, density):
        if not isinstance(density, float):
            raise ValueError("Density must be a float.")
        self.density = density

    def set_viscosity(self, viscosity_type, viscosity_value):
        if not isinstance(viscosity_type, str):
            raise ValueError("Viscosity type must be a string.")
        if viscosity_type not in ["Constant"]:
            raise ValueError("Viscosity type must be 'Constant'.")
        if not isinstance(viscosity_value, float):
            raise ValueError("Viscosity value must be a float.")
        self.viscosity = {"type": viscosity_type, "value": viscosity_value}

    def set_conductivity(self, conductivity):
        if not isinstance(conductivity, float):
            raise ValueError("Conductivity must be a float.")
        self.conductivity = conductivity

    def set_source_term(self, source_term):
        if not isinstance(source_term, float):
            raise ValueError("Source term must be a float.")
        self.source_term = source_term

    def add_mesh(self, mesh):
        if not isinstance(mesh, GeneralMesh):
            raise ValueError("Mesh must be a GeneralMesh object.")
        self.meshes.append(mesh)
        self.faces.extend(list(mesh.faces.keys()))

    def add_bc(self, face, bc_type, value, time_varying=False,
               time_varying_file=None, impose_flux=False, profile=None,
               zero_out_perimeter=True):
        if face not in self.faces:
            raise ValueError("Face not found in mesh.")
        for bc in self.boundary_conditions:
            if bc.name == face:
                raise ValueError("Boundary condition already exists for face.")
        bc = BoundaryCondition()
        bc.set_name(face)
        bc.set_type(bc_type)
        if not time_varying:
            bc.set_value(value)
        else:
            if isinstance(time_varying_file, type(None)):
                raise ValueError("Time varying file must be set.")
            bc.set_time_varying(time_varying)
            bc.set_time_varying_file(time_varying_file)
        bc.impose_flux = impose_flux
        bc.profile = profile
        self.boundary_conditions.append(bc)

    def reassign_bc(self, face, bc_type, value, time_varying=False,
                    time_varying_file=None, impose_flux=False, profile=None):
        if face not in self.faces:
            raise ValueError("Face not found in mesh.")
        bc = BoundaryCondition()
        bc.set_name(face)
        bc.set_type(bc_type)
        if not time_varying:
            bc.set_value(value)
        else:
            if isinstance(time_varying_file, type(None)):
                raise ValueError("Time varying file must be set.")
            bc.set_time_varying(time_varying)
            bc.set_time_varying_file(time_varying_file)
        bc.impose_flux = impose_flux
        bc.profile = profile
        for i, b in enumerate(self.boundary_conditions):
            if b.name == face:
                self.boundary_conditions[i] = bc
                break

    def toxml(self):
        add_equation = self.file.createElement("Add_equation")
        if not isinstance(self.equation_type, type(None)):
            add_equation.setAttribute("type", self.equation_type)
        else:
            raise ValueError("Equation type must be set.")

        if not isinstance(self.coupled, type(None)):
            coupled = self.file.createElement("Coupled")
            if self.coupled:
                coupled.appendChild(self.file.createTextNode("true"))
            else:
                coupled.appendChild(self.file.createTextNode("false"))
            add_equation.appendChild(coupled)

        if not isinstance(self.min_iterations, type(None)):
            min_iterations = self.file.createElement("Min_iterations")
            min_iterations.appendChild(self.file.createTextNode(str(self.min_iterations)))
            add_equation.appendChild(min_iterations)

        if not isinstance(self.max_iterations, type(None)):
            max_iterations = self.file.createElement("Max_iterations")
            max_iterations.appendChild(self.file.createTextNode(str(self.max_iterations)))
            add_equation.appendChild(max_iterations)

        if not isinstance(self.tolerance, type(None)):
            tolerance = self.file.createElement("Tolerance")
            tolerance.appendChild(self.file.createTextNode(str(self.tolerance)))
            add_equation.appendChild(tolerance)

        if not isinstance(self.backflow_stabilization_coefficient, type(None)):
            backflow_stabilization_coefficient = self.file.createElement("Backflow_stabilization_coefficient")
            backflow_stabilization_coefficient.appendChild(self.file.createTextNode(str(self.backflow_stabilization_coefficient)))
            add_equation.appendChild(backflow_stabilization_coefficient)

        if not isinstance(self.density, type(None)):
            density = self.file.createElement("Density")
            density.appendChild(self.file.createTextNode(str(self.density)))
            add_equation.appendChild(density)

        if not isinstance(self.conductivity, type(None)):
            conductivity = self.file.createElement("Conductivity")
            conductivity.appendChild(self.file.createTextNode(str(self.conductivity)))
            add_equation.appendChild(conductivity)

        if not isinstance(self.source_term, type(None)):
            source_term = self.file.createElement("Source_term")
            source_term.appendChild(self.file.createTextNode(str(self.source_term)))
            add_equation.appendChild(source_term)

        if not isinstance(self.viscosity, type(None)):
            viscosity = self.file.createElement("Viscosity")
            viscosity.setAttribute("model", self.viscosity["type"])
            value = self.file.createElement("Value")
            value.appendChild(self.file.createTextNode(str(self.viscosity["value"])))
            viscosity.appendChild(value)
            add_equation.appendChild(viscosity)

        for out in self.outputs:
            if isinstance(out, Output):
                add_equation.appendChild(out.toxml())

        if isinstance(self.alias, Output) and self.alias.output_type == "Alias":
            add_equation.appendChild(self.alias.toxml())

        if isinstance(self.linear_solver, LinearSolver):
            add_equation.appendChild(self.linear_solver.toxml())

        for boundary_condition in self.boundary_conditions:
            add_equation.appendChild(boundary_condition.toxml())

        return add_equation



import pytest
import numpy as np

# If your compiled module is named something else, replace the import accordingly:
from svv.domain.routines.c_sample import (
    pick_from_tetrahedron,
    pick_from_triangle,
    pick_from_line
)

def test_pick_from_line():
    """
    A line can be defined by two 3D points. For example:
        point A = [0, 0, 0]
        point B = [1, 1, 1]
    If the barycentric coordinate for B is x, then for A is (1 - x).
    The resulting point is (1 - x)*A + x*B = [x, x, x].
    """
    # We'll test a single sample (i=1)
    simplices = np.array([
        [[0.0, 0.0, 0.0],  # Corner A
         [1.0, 1.0, 1.0]]  # Corner B
    ], dtype=np.float64).reshape(1, 2, 3)

    # rdx is shape (i, 2, 1). We'll choose x=0.2, so rdx[0,1,0] = 0.2, rdx[0,0,0] is ignored initially
    rdx = np.array([
        [[0.0],
         [0.2]]
    ], dtype=np.float64).reshape(1, 2, 1)

    # Expected result => [0.2, 0.2, 0.2]
    result = pick_from_line(simplices, rdx)
    expected = np.array([[0.2, 0.2, 0.2]], dtype=np.float64)
    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_pick_from_triangle():
    """
    A triangle can be defined by 3 points in 3D. For example:
        A = [0, 0, 0]
        B = [1, 0, 0]
        C = [0, 1, 0]
    We'll pick barycentric coordinates (r2, r3). The first coordinate is r1 = 1 - r2 - r3.
    The resulting point is r1*A + r2*B + r3*C.
    """
    simplices = np.array([
        [[0.0, 0.0, 0.0],  # A
         [1.0, 0.0, 0.0],  # B
         [0.0, 1.0, 0.0]]  # C
    ], dtype=np.float64).reshape(1, 3, 3)

    # rdx shape (1, 3, 1).
    # Suppose r2=0.2, r3=0.3 => r1=0.5 => final point = [0.2, 0.3, 0.0]
    rdx = np.array([
        [[0.0],
         [0.2],
         [0.3]]
    ], dtype=np.float64).reshape(1, 3, 1)

    # Expected => [0.2, 0.3, 0.0]
    result = pick_from_triangle(simplices, rdx)
    expected = np.array([[0.2, 0.3, 0.0]], dtype=np.float64)
    assert np.allclose(result, expected), f"Expected {expected}, got {result}"


def test_pick_from_tetrahedron():
    """
    A tetrahedron can be defined by 4 points in 3D. For example:
        A = [0, 0, 0]
        B = [1, 0, 0]
        C = [0, 1, 0]
        D = [0, 0, 1]
    We'll specify barycentric coords (r2, r3, r4). The function
    modifies them if sums exceed 1, but let's choose a case that
    doesn't trigger many conditionals, for clarity.
    Suppose r2=0.2, r3=0.2, r4=0.3 => r1=0.3 => final point = [0.2, 0.2, 0.3].
    """
    simplices = np.array([[
        [0.0, 0.0, 0.0],  # A
        [1.0, 0.0, 0.0],  # B
        [0.0, 1.0, 0.0],  # C
        [0.0, 0.0, 1.0]   # D
    ]], dtype=np.float64)  # shape (1, 4, 3)

    # shape (1, 4, 1). We'll manually set r2=0.2, r3=0.2, r4=0.3 => r1=0
    # The function itself sets r1 = 1 - sum(r2..r4), so we can initialize r1=0.0 for convenience.
    rdx = np.array([
        [[0.0],
         [0.2],
         [0.2],
         [0.3]]
    ], dtype=np.float64).reshape(1, 4, 1)

    result = pick_from_tetrahedron(simplices, rdx)
    # We expect the final barycentric coords to remain r1=0.3, r2=0.2, r3=0.2, r4=0.3 =>
    # => [0.2, 0.2, 0.3]
    expected = np.array([[0.2, 0.2, 0.3]], dtype=np.float64)
    assert np.allclose(result, expected), f"Expected {expected}, got {result}"



import pyvista
from tqdm import trange

def show(tree, color='red', plot_domain=False, return_plotter=False, **kwargs):
    """
    Visualize a synthetic vascular tree using PyVista.

    This function creates a 3D visualization of a vascular tree, where each vessel is represented
    as a cylinder. The tree's data is iterated over to generate the cylinders based on their
    geometrical properties such as center, direction, radius, and length.

    Parameters
    ----------
    tree : object
        The synthetic vascular tree data structure. This object should have a `data` attribute,
        which is an array where each row represents a vessel. The `data` array is expected to have
        columns corresponding to the start and end points of the vessel (first three and next three
        columns, respectively). Additionally, the tree should provide access to methods like
        `get('w_basis', i)`, `get('radius', i)`, and `get('length', i)` that return the direction
        vector, radius, and length of each vessel, respectively.

    color : str, optional
        The color of the vessels in the visualization. Default is 'red'.

    plot_domain : bool, optional
        If True, the boundary of the domain in which the tree exists is also plotted with a
        semi-transparent grey mesh. Default is False.

    return_plotter : bool, optional
        If True, the PyVista `Plotter` object is returned instead of displaying the plot immediately.
        This can be useful for further customization of the plot outside of this function. Default
        is False.

    **kwargs : dict, optional
        Additional keyword arguments passed to the PyVista `Plotter` initializer. These can be used
        to control the appearance of the plot, such as the background color or window size.

    Returns
    -------
    plotter : pyvista.Plotter, optional
        If `return_plotter` is True, the PyVista `Plotter` object is returned. This allows further
        modifications to the plot before it is displayed.

    Notes
    -----
    The function uses a progress bar from `tqdm` to provide feedback during the plotting process.
    This can be useful when dealing with large trees, as it shows the progress of building the plot
    in real time.

    Examples
    --------

    .. code-block:: python

        >>> from svtoolkit import Tree, Domain
        >>> from svtoolkit.visualize.tree import show
        >>> import pyvista as pv
        >>> cube = Domain(pv.Cube()) # Create a cube domain from a PyVista mesh
        >>> cube.create() # Create the domain
        >>> cube.solve() # Solve the implicit functions of the domain
        >>> cube.build() # Build the domain
        >>> t = Tree() # Create a tree object
        >>> t.set_domain(cube) # Set the domain for the tree
        >>> t.set_root() # Set the root vessel
        >>> t.n_add(10) # Add 10 vessels to the tree
        >>> show(t, plot_domain=True) # Visualize the tree with the domain boundary

    """
    plotter = pyvista.Plotter(**kwargs)
    for i in trange(tree.data.shape[0], desc='Building plot', unit='vessel', leave=False):
        center = (tree.data[i, 0:3] + tree.data[i, 3:6]) / 2
        direction = tree.data.get('w_basis', i)
        radius = tree.data.get('radius', i)
        length = tree.data.get('length', i)
        vessel = pyvista.Cylinder(center=center, direction=direction, radius=radius, height=length)
        plotter.add_mesh(vessel, color=color)
    if plot_domain:
        plotter.add_mesh(tree.domain.boundary, color='grey', opacity=0.25)
    if return_plotter:
        return plotter
    else:
        plotter.show()




"""
Remeshing utility based on MMG executables.

This module provides functions to remesh surfaces and volumes using MMG, a software for remeshing.

Functions
---------
- remesh_surface_2d : Remeshes 2D surface boundaries using MMG2D.
- remesh_surface : Remeshes 3D surfaces using MMGS.
- remesh_volume : Remeshes 3D volumes using MMG3D.
- add_required : Adds required triangles to a .mesh file.
- clean_medit : Cleans up a .mesh file by removing unsupported keywords.
"""


from xml.dom import minidom


class LinearSolverBase:
    def __init__(self):
        self.solver_type = "NS"
        self.absolute_tolerance = 1.0e-10
        self.tolerance = 0.5
        self.max_iterations = 1000
        self.preconditioner = None
        self.ns_cg_max_iterations = 1000
        self.ns_cg_tolerance = 1.0e-2
        self.ns_gm_max_iterations = 1000
        self.ns_gm_tolerance = 1.0e-2
        self.use_trilinos_for_assembly = False
        self.krylov_space_dimension = 50

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return str(vars(self))

    def __eq__(self, other):
        check = ["solver_type", "absolute_tolerance", "tolerance", "max_iterations", "preconditioner",
                 "ns_cg_max_iterations", "ns_cg_tolerance", "ns_gm_max_iterations", "ns_gm_tolerance",
                 "use_trilinos_for_assembly"]
        attributes = vars(self)
        other_attributes = vars(other)
        return all(other_attributes[key] == attributes[key] for key in check)

    def __ne__(self, other):
        return not self.__eq__(other)


class LinearSolver(LinearSolverBase):
    def __init__(self):
        super().__init__()
        self.file = minidom.Document()

    def __str__(self):
        return self.toxml().toprettyxml()

    def __repr__(self):
        return self.toxml().toprettyxml()

    def set_type(self, solver_type):
        """
        Set the linear solver type. Options are 'CG', 'GMRES', 'NS', 'BICGS'.
        Note that not all linear solvers and preconditioners are compatible.
        """
        if solver_type not in ["CG", "GMRES", "NS", "BICGS"]:
            raise ValueError("Solver type must be 'CG' or 'GMRES'.")
        self.solver_type = solver_type

    def toxml(self):
        ls = self.file.createElement("LS")
        if not isinstance(self.solver_type, type(None)):
            ls.setAttribute("type", self.solver_type)
        else:
            raise ValueError("Solver type must be set.")

        if not isinstance(self.absolute_tolerance, type(None)):
            absolute_tolerance = self.file.createElement("Absolute_tolerance")
            absolute_tolerance.appendChild(self.file.createTextNode(str(self.absolute_tolerance)))
            ls.appendChild(absolute_tolerance)

        if not isinstance(self.tolerance, type(None)):
            tolerance = self.file.createElement("Tolerance")
            tolerance.appendChild(self.file.createTextNode(str(self.tolerance)))
            ls.appendChild(tolerance)

        if not isinstance(self.max_iterations, type(None)):
            max_iterations = self.file.createElement("Max_iterations")
            max_iterations.appendChild(self.file.createTextNode(str(self.max_iterations)))
            ls.appendChild(max_iterations)

        if not isinstance(self.preconditioner, type(None)):
            preconditioner = self.file.createElement("Preconditioner")
            preconditioner.appendChild(self.file.createTextNode(self.preconditioner))
            ls.appendChild(preconditioner)

        if not isinstance(self.ns_cg_max_iterations, type(None)):
            ns_cg_max_iterations = self.file.createElement("NS_CG_max_iterations")
            ns_cg_max_iterations.appendChild(self.file.createTextNode(str(self.ns_cg_max_iterations)))
            ls.appendChild(ns_cg_max_iterations)

        if not isinstance(self.ns_cg_tolerance, type(None)):
            ns_cg_tolerance = self.file.createElement("NS_CG_tolerance")
            ns_cg_tolerance.appendChild(self.file.createTextNode(str(self.ns_cg_tolerance)))
            ls.appendChild(ns_cg_tolerance)

        if not isinstance(self.ns_gm_max_iterations, type(None)):
            ns_gm_max_iterations = self.file.createElement("NS_GM_max_iterations")
            ns_gm_max_iterations.appendChild(self.file.createTextNode(str(self.ns_gm_max_iterations)))
            ls.appendChild(ns_gm_max_iterations)

        if not isinstance(self.ns_gm_tolerance, type(None)):
            ns_gm_tolerance = self.file.createElement("NS_GM_tolerance")
            ns_gm_tolerance.appendChild(self.file.createTextNode(str(self.ns_gm_tolerance)))
            ls.appendChild(ns_gm_tolerance)

        if self.use_trilinos_for_assembly:
            use_trilinos_for_assembly = self.file.createElement("Use_trilinos_for_assembly")
            use_trilinos_for_assembly.appendChild(self.file.createTextNode("true"))
            ls.appendChild(use_trilinos_for_assembly)

        if not isinstance(self.krylov_space_dimension, type(None)):
            krylov_space_dimension = self.file.createElement("Krylov_space_dimension")
            krylov_space_dimension.appendChild(self.file.createTextNode(str(self.krylov_space_dimension)))
            ls.appendChild(krylov_space_dimension)
        return ls



# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import json
import pdb
import numpy as np

from .parameters import OutflowBoundaryConditionType


def write_0d_solver_file(mesh, params, model):
    """
    Generate 0d solver input file (.json)
    """
    # create input dictionary
    inp = {'simulation_parameters': {},
           'boundary_conditions': [],
           'junctions': [],
           'vessels': []}

    # general
    inp['simulation_parameters']['model_name'] = params.model_name

    # time
    dt = params.time_step
    n_step = params.num_time_steps
    t_cycle = mesh.inflow_data[-1][0]
    inp['simulation_parameters']['number_of_time_pts_per_cardiac_cycle'] = int(t_cycle / dt)
    inp['simulation_parameters']['number_of_cardiac_cycles'] = int(n_step / t_cycle * dt) + 1

    # fluid
    inp['simulation_parameters']['density'] = params.density
    inp['simulation_parameters']['viscosity'] = params.viscosity

    # vessels
    for (branch, ids), lengths in zip(mesh.cell_data['id'].items(), mesh.cell_data['length'].values()):
        for i, (j, length) in enumerate(zip(ids, lengths)):
            vessel = {'vessel_id': int(j),
                      'vessel_name': 'branch' + str(branch) + '_seg' + str(i),
                      'vessel_length': length,
                      'zero_d_element_type': 'BloodVessel',
                      'zero_d_element_values': {}}
            # zerod values
            for zerod in model['branches'].keys():
                vessel['zero_d_element_values'][zerod] = model['branches'][zerod][branch][i]
            # inlet bc
            if branch == 0 and i == 0:
                vessel['boundary_conditions'] = {'inlet': 'INFLOW'}
            # outlet bc
            if j in mesh.terminal:
                bc_name = list(mesh.outlet_face_names_index.keys())[mesh.terminal.index(j)]
                bc_str = mesh.bc_type[bc_name].upper() + '_' + str(mesh.outlet_face_names_index[bc_name])
                vessel['boundary_conditions'] = {'outlet': bc_str}
            inp['vessels'] += [vessel]

    # junctions
    for i, j_branches in enumerate(mesh.seg_connectivity):
        # general junction properties
        junction = {'junction_name': 'J' + str(i),
                    'junction_type': 'internal_junction',
                    'inlet_vessels': [int(j_branches[0])],
                    'outlet_vessels': []}
        for j in j_branches[1:]:
            junction['outlet_vessels'] += [int(j)]
        # branching vessels
        if len(j_branches) > 2:
            junction['junction_type'] = 'BloodVesselJunction'
            junction['tangents'] = mesh.junctions[i]['tangents']
            for n in ['areas', 'lengths']:
                junction[n] = [float(v) for v in mesh.junctions[i][n]]

        inp['junctions'] += [junction]

    # inlet bc
    inflow_q = []
    inflow_t = []
    for value in mesh.inflow_data:
        inflow_t += [value.time]
        inflow_q += [value.flow]
    inflow = {'bc_name': 'INFLOW',
              'bc_type': 'FLOW',
              'bc_values': {'t': inflow_t, 'Q': inflow_q}}
    inp['boundary_conditions'] += [inflow]

    # outlet bcs
    for bc_name, i in mesh.outlet_face_names_index.items():
        bc_type = mesh.bc_type[bc_name]
        bc_str = mesh.bc_type[bc_name].upper() + '_' + str(i)
        bc_val = mesh.bc_map[bc_name]

        outflow = {'bc_name': bc_str,
                   'bc_type': bc_type.upper(),
                   'bc_values': {}}
        if bc_type == OutflowBoundaryConditionType.RCR:
            seq = ['Rp', 'C', 'Rd', 'Pd']
            for name, val in zip(seq, bc_val):
                outflow['bc_values'][name] = float(val)
        elif bc_type == OutflowBoundaryConditionType.RESISTANCE:
            seq = ['R', 'Pd']
            for name, val in zip(seq, bc_val):
                outflow['bc_values'][name] = float(val)
        elif bc_type == OutflowBoundaryConditionType.CORONARY:
            for name, val in bc_val['var'].items():
                outflow['bc_values'][name] = val
            outflow['bc_values']['t'] = bc_val['time']
            outflow['bc_values']['Pim'] = bc_val['pressure']
            outflow['bc_values']['Pv'] = 0.0
        inp['boundary_conditions'] += [outflow]

    # write to file
    file_name = os.path.join(params.output_directory, params.solver_output_file)
    with open(file_name, 'w') as file:
        json.dump(inp, file, indent=4, sort_keys=True)



import numpy as np
import vtk
import pyvista as pv
from vtk.util import numpy_support
from skimage.measure import marching_cubes

def descritize(domain, **kwargs):
    """
    This function evaluates the implicit representation
    of the domain and returns a discrete representation
    as a 2D/3D mesh.

    Parameters
    ----------
    domain : svtoolkit.Domain object
        This is the domain object to be discretized.

    Return
    ------
    mesh : vtk.vtkPolyData or vtk.vtkUnstructuredGrid
        This is the discrete representation of the domain.
        If the domain is 2D, then a vtkPolyData object is
        returned. If the domain is 3D, then a vtkUnstructuredGrid
        object is returned.

    Notes
    -----
    """
    buffer = kwargs.get("buffer", 1.0)
    resolution = kwargs.get("resolution", 20)
    mins = np.min(domain.points, axis=0)
    maxs = np.max(domain.points, axis=0)
    ranges = maxs - mins
    origin = (maxs + mins)/2
    buffers = (ranges * buffer) / 2
    scale = (ranges + 2*buffers)
    if domain.d == 2:
        mesh = threshold_2d(domain.evaluate, resolution, origin, scale, 0, -1)
    elif domain.d == 3:
        mesh = threshold_3d(domain.evaluate, resolution, origin, scale, 0, -1)
    else:
        raise ValueError("Only 2D and 3D domains are supported.")
    return mesh


def boundaries(domain, **kwargs):
    pass


def threshold_2d(function, resolution, origin, scale, upper, lower):
    """
    This function uses thresholding techniques to
    discretize a 2D subdomain of a defined 2D
    implicit domain.

    Parameters
    ----------
    function : callable
        This is the function that defines the implicit
        domain.
    resolution : int
        This is the number of points to discretize the
        two nominal dimensions of the domain.
    origin : array_like
        This is the origin of the domain.
    scale : array_like
        This is the scale of the domain.
    upper : float
        This is the upper threshold value.
    lower : float
        This is the lower threshold value.

    Returns
    -------
    mesh : vtkPolyData
        This is the discrete mesh of the 2D domain.
        The vtkPolyData object retains the scalar
        values of the function evaluated at each
        point. The scalar values are stored in the
        vtkPolyData point data as array named
        "ImplicitFunctionValue".

    Notes
    -----

    """
    # Build a plane source to evaluate the function
    plane = vtk.vtkPlaneSource()
    plane.SetXResolution(resolution)
    plane.SetYResolution(resolution)
    plane.Update()
    # Define the transform to scale the plane
    transform = vtk.vtkTransform()
    transform.Scale(scale[0], scale[1], 1)
    transform.Translate(origin[0], origin[1], origin[2])
    # Apply the transform to the plane
    transform_filter = vtk.vtkTransformPolyDataFilter()
    transform_filter.SetInputData(plane.GetOutput())
    transform_filter.SetTransform(transform)
    transform_filter.Update()
    plane = transform_filter.GetOutput()
    # Evaluate the function on the plane
    points = numpy_support.vtk_to_numpy(transform_filter.GetOutput().GetPoints().GetData())[:, :2]
    values = function(points).flatten()
    values = numpy_support.numpy_to_vtk(values)
    values.SetName("ImplicitFunctionValue")
    # Set implicit values to the plane
    plane.GetPointData().SetScalars(values)
    # Create an UnstructuredGrid from the PolyData
    polydata_to_unstructured_grid = vtk.vtkAppendFilter()
    polydata_to_unstructured_grid.AddInputData(plane)
    polydata_to_unstructured_grid.Update()
    unstructured_grid = vtk.vtkUnstructuredGrid()
    unstructured_grid.ShallowCopy(polydata_to_unstructured_grid.GetOutput())
    # Threshold the plane
    threshold = vtk.vtkThreshold()
    threshold.SetInputData(unstructured_grid)
    threshold.ThresholdBetween(lower, upper)
    threshold.SetInputArrayToProcess(0, 0, 0, vtk.vtkDataObject.FIELD_ASSOCIATION_POINTS, "ImplicitFunctionValue")
    threshold.Update()
    # Obtain the mesh
    geometry_filter = vtk.vtkGeometryFilter()
    geometry_filter.SetInputData(threshold.GetOutput())
    geometry_filter.Update()
    mesh = geometry_filter.GetOutput()
    return mesh


def threshold_3d(function, resolution, origin, scale, upper, lower):
    """
    This function uses thresholding techniques to
    discretize a 3D subdomain of a defined 3D implicit
    domain.

    Parameters
    ----------
    function : callable
        This is the function that defines the implicit
        domain.
    resolution : int
        This is the number of points to discretize the
        three nominal dimensions of the domain.
    origin : array_like
        This is the origin of the domain.
    scale : array_like
        This is the scale of the domain.
    upper : float
        This is the upper threshold value.
    lower : float
        This is the lower threshold value.

    Returns
    -------
    mesh : vtkUnstructuredGrid
        This is the discrete mesh of the 3D domain.
        The unstructured grid also retains the implicit
        values used to threshold the domain. These data
        are stored in the point data of the unstructured
        grid as a scalar array named "ImplicitFunctionValue".

    Notes
    -----

    """
    # Build the structured grid for thresholding
    uniform_grid = vtk.vtkUniformGrid()
    uniform_grid.SetDimensions(resolution, resolution, resolution)
    # Define the transform to scale the grid
    transform = vtk.vtkTransform()
    transform.Scale(scale[0], scale[1], scale[2])
    transform.Translate(origin[0], origin[1], origin[2])
    # Apply the transform to the grid
    transform_filter = vtk.vtkTransformFilter()
    transform_filter.SetInputData(uniform_grid)
    transform_filter.SetTransform(transform)
    transform_filter.Update()
    uniform_grid = transform_filter.GetOutput()
    # Evaluate the function on the grid
    points = numpy_support.vtk_to_numpy(uniform_grid.GetPoints().GetData())
    values = function(points).flatten()
    values = numpy_support.numpy_to_vtk(values)
    values.SetName("ImplicitFunctionValue")
    # Set implicit values to the grid
    uniform_grid.GetPointData().SetScalars(values)
    # Transform the grid to an unstructured grid
    append_filter = vtk.vtkAppendFilter()
    append_filter.AddInputData(uniform_grid)
    append_filter.Update()
    unstructured_grid = vtk.vtkUnstructuredGrid()
    unstructured_grid.ShallowCopy(append_filter.GetOutput())
    # Threshold the grid
    threshold = vtk.vtkThreshold()
    threshold.SetInputData(unstructured_grid)
    threshold.ThresholdBetween(lower, upper)
    threshold.SetInputArrayToProcess(0, 0, 0, vtk.vtkDataObject.FIELD_ASSOCIATION_POINTS, "ImplicitFunctionValue")
    threshold.Update()
    # Obtain the mesh
    mesh = threshold.GetOutput()
    return mesh

def marching_squares(function, grid, origin, value=0.0):
    """
    This function uses the marching squares algorithm
    to discretize to determine the boundary of a 2D
    domain.

    Parameters
    ----------
    function : callable
        This is the implicit function that defines the
        domain.
    grid : numpy.ndarray
        This is an array of points that define the grid
        on which the implicit function is evaluated.

    Returns
    -------
    mesh : vtkPolyData
        This is the mesh that represents the 1D boundary
        along the 2D domain.
    """
    dimension_0_spacing = np.diff(grid[0][0, :])[0]
    dimension_1_spacing = np.diff(grid[1][:, 0])[0]
    dimension_0 = grid[0].flatten()
    dimension_1 = grid[1].flatten()
    dimensions = np.vstack((dimension_0, dimension_1)).T
    values = function(dimensions).flatten()
    image_data = vtk.vtkImageData()
    image_data.SetDimensions(grid[0].shape[0], grid[0].shape[1], 1)
    image_data.SetSpacing(dimension_0_spacing, dimension_1_spacing, 1)
    image_data.SetOrigin(origin[0], origin[1], 0)
    image_data.AllocateScalars(vtk.VTK_DOUBLE, 1)
    vtk_values = numpy_support.numpy_to_vtk(values, deep=True, array_type=vtk.VTK_DOUBLE)
    image_data.GetPointData().SetScalars(vtk_values)
    # Create the 2D filter
    marching_squares_filter = vtk.vtkMarchingSquares()
    marching_squares_filter.SetInputData(image_data)
    marching_squares_filter.SetValue(0, value)
    # Obtain the mesh
    marching_squares_filter.Update()
    mesh = marching_squares_filter.GetOutput()
    return mesh

#def marching_cubes(function, grid, origin, value=0.0):
#    pass

def contour(function, points, resolution, value=0.0, buffer=1.5):
    n_dim = points.shape[1]
    if n_dim == 2:
        bounds = [np.min(points[:, 0]), np.max(points[:, 0]),
                  np.min(points[:, 1]), np.max(points[:, 1])]
        origin = [(bounds[1]+bounds[0])/2,
                  (bounds[3]+bounds[2])/2,
                  0]
        mesh = pv.Plane(i_resolution=resolution, j_resolution=resolution)
        mesh.points[:, 0] *= (bounds[1] - bounds[0])*buffer
        mesh.points[:, 1] *= (bounds[3] - bounds[2])*buffer
        mesh.points += origin
        mesh.point_data["values"] = function(mesh.points[:, :2])
        boundary = mesh.contour([value])
    elif n_dim == 3:
        bounds = [np.min(points, axis=0), np.max(points, axis=0)]
        bounds_range = np.abs(bounds[1] - bounds[0])*buffer
        x = np.linspace(bounds[0][0] - bounds_range[0], bounds[1][0] + bounds_range[0], resolution)
        y = np.linspace(bounds[0][1] - bounds_range[1], bounds[1][1] + bounds_range[1], resolution)
        z = np.linspace(bounds[0][2] - bounds_range[2], bounds[1][2] + bounds_range[2], resolution)
        x, y, z = np.meshgrid(x, y, z, indexing="ij")
        strucutrured_grid = pv.StructuredGrid(x, y, z)
        strucutrured_grid.point_data["values"] = function(strucutrured_grid.points)
        boundary = strucutrured_grid.contour([value])
        mesh = strucutrured_grid
        """
        origin = (bounds[1] + bounds[0]) / 2
        ranges = (bounds[1] - bounds[0])*buffer
        spacing = ranges / resolution
        dimensions = [resolution, resolution, resolution]
        #origin = np.array(origin) - np.array(ranges)/2
        mesh = pv.ImageData(spacing=spacing, dimensions=dimensions, origin=origin)
        grid = np.mgrid[origin[0] - ranges[0]/2:origin[0] + ranges[0]/2:resolution*1j,
                        origin[1] - ranges[1]/2:origin[1] + ranges[1]/2:resolution*1j,
                        origin[2] - ranges[2]/2:origin[2] + ranges[2]/2:resolution*1j]
        grid = grid.reshape(3, -1).T
        values = function(grid)
        values = values.reshape(resolution, resolution, resolution)
        mesh.point_data["values"] = values.flatten()
        #boundary = mesh.contour([value], method="flying_edges")
        verts, faces, normals, _ = marching_cubes(values, level=value, spacing=spacing,
                                                  allow_degenerate=False)
        offset = (verts.max(axis=0) + verts.min(axis=0)) / 2
        shift = (origin - offset)
        verts = (verts) + shift
        face_data = np.zeros((faces.shape[0], 4), dtype=int)
        face_data[:, 0] = 3
        face_data[:, 1:] = faces
        boundary = pv.PolyData(verts, face_data)
        boundary.point_data["values"] = function(verts).flatten()
        """
    else:
        boundary = None
        mesh = None
    return boundary, mesh





import numpy

from scipy.spatial import cKDTree
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import shortest_path


def geodesic_constructor(domain, **kwargs):
    """
    Construct a general geodesic function solver for a given domain.

    Parameters
    ----------
    domain : svtoolkit.domain.Domain
        The domain object that defines the spatial region in which vascular
        trees are generated.
    kwargs : dict
        Additional keyword arguments.
        Keyword arguments include:


    Returns
    -------
    get_geodesic : function
        A function that computes the geodesic path between two points.
    """
    idx = [[0, 1], [1, 2], [2, 0], [0, 3], [3, 1], [2, 3]]
    tetra = domain.mesh.cells.reshape(-1, 5)[:, 1:]
    lengths = []
    nodes = []
    added_nodes = set([])
    tetra_node_tree = cKDTree(domain.mesh.points)
    for i in range(tetra.shape[0]):
        tet = tetra[i, :]
        for cx in idx:
            if tuple([tet[cx[0]], tet[cx[1]]]) in added_nodes:
                continue
            added_nodes.add(tuple([tet[cx[0]], tet[cx[1]]]))
            added_nodes.add(tuple([tet[cx[1]], tet[cx[0]]]))
            nodes.append([tet[cx[0]], tet[cx[1]]])
            nodes.append([tet[cx[1]], tet[cx[0]]])
            length = numpy.linalg.norm(domain.mesh.points[tet[cx[0]], :] - domain.mesh.points[tet[cx[1]], :])
            lengths.append(length)
            lengths.append(length)
    M = numpy.array(nodes)
    L = numpy.array(lengths)
    graph = csr_matrix((L, (M[:, 0], M[:, 1])), shape=(numpy.max(M[:, 0]) + 1, numpy.max(M[:, 1]) + 1))

    def get_path(start, end, graph=graph, shortest_path=shortest_path):
        dist, pred = shortest_path(csgraph=graph, directed=False, indices=start, return_predecessors=True)
        path = [end]
        dists = []
        k = end
        while pred[k] != -9999:
            path.append(pred[k])
            dists.append(dist[k])
            k = pred[k]
        path = path[::-1]
        dists = dists[::-1]
        lines = []
        for i in range(len(path) - 1):
            lines.append([path[i], path[i + 1]])
        return path, dists, lines

    def get_geodesic(start, end, tetra_node_tree=tetra_node_tree, get_path=get_path):
        """
        Get the geodesic path between two points

        Parameters
        ----------
        start : numpy.ndarray
            The starting point.
        end : numpy.ndarray
            The ending point.
        tetra_node_tree : scipy.spatial.cKDTree
            The tree of the tetrahedral nodes. This parameter is partially applied and
            should not be modified.
        get_path : function
            The function that computes the path between two nodes. This parameter is
            partially applied and should not be modified.

        Returns
        -------
        path : list
            The list of nodes (mesh indices) in the path.
        dists : list
            The list of distances between nodes.
        lines : list
            The list of lines between nodes.
        """
        ind = tetra_node_tree.query(start)[1]
        jnd = tetra_node_tree.query(end)[1]
        path, dists, lines = get_path(ind, jnd)
        return path, dists, lines
    return get_geodesic




from abc import ABC

import numpy as np

from svv.forest.connect.base import BaseCurve


class NURBSCurve(BaseCurve, ABC):
    """
    A concrete implementation of a Non-Uniform Rational B-Spline (NURBS) curve.

    This class inherits from the abstract base Curve. It satisfies the required
    interface by implementing (or stubbing) all abstract methods.

    Parameters
    ----------
    control_points : (N, d) array-like
        The N control points, each of dimension d (d=2 or d=3 typically).
    weights : (N,) array-like or None, optional
        Weights associated with each control point. Must match length of control_points.
        Defaults to all 1's (uniform).
    knot_vector : (M,) array-like or None, optional
        The non-decreasing knot vector. Typically, M = N + degree + 1 for a clamped spline.
        If not provided, a default clamped knot vector is used.
    degree : int or None, optional
        The polynomial degree of the B-spline basis (e.g., 2 for quadratic, 3 for cubic).
        Defaults to `min(len(control_points) - 1, 3)`.

    Notes
    -----
    - This minimal class focuses on the evaluate() method, which takes parameter t in [0,1]
      and maps it to the relevant knot span [knot_vector[degree], knot_vector[-(degree+1)]].
    - The derivative, roc, torsion, and arc_length methods currently raise NotImplementedError.
      You may implement them with analytic or numeric approaches if needed.
    """

    def __init__(self, control_points, weights=None, knot_vector=None, degree=None):
        # Convert control_points to NumPy array
        self.control_points = np.array(control_points, dtype=float)

        # 1) Default weights: all 1s if not given
        if weights is None:
            self.weights = np.ones(len(self.control_points))
        else:
            self.weights = np.array(weights, dtype=float)

        # 2) Default degree: cubic or the maximum permissible (len - 1) if smaller
        if degree is None:
            self.degree = min(len(self.control_points) - 1, 3)
        else:
            self.degree = degree

        # 3) Default knot vector: clamped with uniform spacing in the interior
        #
        #    A "clamped" knot vector has multiplicities of (degree+1) at the
        #    beginning and end. For instance, if degree=3 and we have 5 control
        #    points, the vector might look like:
        #    [0, 0, 0, 0, 0.5, 1, 1, 1, 1]
        if knot_vector is None:
            num_ctrl_pts = len(self.control_points)
            n_knots = num_ctrl_pts + self.degree + 1

            # interior knots are uniformly spaced. For example, if degree=3 and
            # we have 5 points, we will have 9 knots total. The first 4 are 0,
            # the last 4 are 1, and there's 1 interior knot which is 0.5.
            # More generally, we have (n_knots - 2*(degree+1)) interior knots.
            n_interior = n_knots - 2 * (self.degree + 1)
            # If there are no interior knots, n_interior will be 0 and
            # np.linspace(0,1,0) is an empty array.

            # build the clamped knot vector
            self.knot_vector = np.concatenate([
                np.zeros(self.degree),                  # left clamped
                np.linspace(0, 1, n_interior + 2)[1:-1], # interior
                np.ones(self.degree)                    # right clamped
            ]) if n_interior >= 0 else None

            # Fallback if we somehow don't have enough points for the above approach:
            if self.knot_vector is None or len(self.knot_vector) != n_knots:
                # simpler fallback: just replicate the standard approach
                self.knot_vector = np.concatenate((
                    np.zeros(self.degree + 1),
                    np.linspace(0, 1, n_knots - 2 * (self.degree + 1)),
                    np.ones(self.degree + 1)
                ))
        else:
            self.knot_vector = np.array(knot_vector, dtype=float)

        # Basic validations
        if len(self.control_points) != len(self.weights):
            raise ValueError("control_points and weights must have the same length.")

        num_ctrl_pts = len(self.control_points)
        if len(self.knot_vector) != num_ctrl_pts + self.degree + 1:
            raise ValueError(
                "knot_vector length must be (num_control_points + degree + 1)."
            )

        if not np.all(self.knot_vector[:-1] <= self.knot_vector[1:]):
            raise ValueError("knot_vector must be non-decreasing.")

        # Precompute the valid param domain from the knot vector
        self._t_min = self.knot_vector[self.degree]
        self._t_max = self.knot_vector[-(self.degree + 1)]

    @property
    def dimension(self):
        """Return the dimension (2 or 3) of the control points."""
        return self.control_points.shape[1]

    def evaluate(self, t_values):
        """
        Evaluate the NURBS curve at the specified parameter values in [0,1].

        Parameters
        ----------
        t_values : float or array-like
            Parametric value(s) in [0, 1].

        Returns
        -------
        np.ndarray
            Evaluated points on the NURBS curve. Shape (len(t_values), dimension).

        Notes
        -----
        - We linearly map each t from [0,1] to the physical domain
          [self._t_min, self._t_max] before evaluating the NURBS basis.
        - The NURBS curve is given by:

            C(t) = (Σ ( N_{i,p}(t) * w_i * P_i )) / (Σ ( N_{i,p}(t) * w_i )),

          where N_{i,p}(t) is the B-spline basis function of degree p using
          the Cox–de Boor recursion, and w_i are the weights.
        """
        t_values = np.atleast_1d(t_values).astype(float)

        # Map [0,1] -> [t_min, t_max]
        physical_ts = self._t_min + t_values * (self._t_max - self._t_min)

        # We'll evaluate at each t in physical_ts
        points_out = np.zeros((len(physical_ts), self.dimension))

        for idx, t_physical in enumerate(physical_ts):
            # Compute numerator and denominator of the NURBS equation
            numerator = np.zeros(self.dimension, dtype=float)
            denominator = 0.0
            for i in range(len(self.control_points)):
                Ni = self._basis_function(i, self.degree, t_physical, self.knot_vector)
                w_i = self.weights[i]
                numerator += Ni * w_i * self.control_points[i]
                denominator += Ni * w_i

            # Safeguard if denominator is near zero
            if abs(denominator) < 1e-14:
                # Typically occurs if t is out of range or at extremes.
                # We'll clamp to first or last control point in that case.
                if t_physical <= self._t_min:
                    points_out[idx] = self.control_points[0]
                else:
                    points_out[idx] = self.control_points[-1]
            else:
                points_out[idx] = numerator / denominator

        return points_out

    def derivative(self, t_values, order=1):
        """
        Compute the nth derivative of the NURBS curve.
        Currently unimplemented.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("NURBS derivative not implemented yet.")

    def roc(self, t_values):
        """
        Radius of curvature not implemented for NURBS yet.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("NURBS radius of curvature not implemented yet.")

    def torsion(self, t_values):
        """
        Torsion not implemented for NURBS yet.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("NURBS torsion not implemented yet.")

    def arc_length(self, t_start=0.0, t_end=1.0, num_points=100):
        """
        Approximate the arc length. Currently unimplemented.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("NURBS arc length not implemented yet.")

    def transform(self, matrix):
        """
        Apply a transformation matrix to the control points.
        For now, raise NotImplementedError or implement if desired.
        """
        raise NotImplementedError("NURBS transform not implemented yet.")

    # ------------------------------------------------------------
    # Cox–de Boor recursion for the B-spline basis N_{i,p}(t).
    # ------------------------------------------------------------
    def _basis_function(self, i, p, t, knots):
        """
        Cox–de Boor recursive definition of B-spline basis function N_{i,p}(t).

        Parameters
        ----------
        i : int
            Index of the basis function.
        p : int
            Degree of the basis function.
        t : float
            Parameter to evaluate at.
        knots : array-like
            Knot vector.

        Returns
        -------
        float
            The value of the B-spline basis function N_{i,p}(t).
        """
        # Base case p=0
        if p == 0:
            return 1.0 if (knots[i] <= t < knots[i+1]) else 0.0

        denom1 = knots[i+p] - knots[i]
        denom2 = knots[i+p+1] - knots[i+1]

        term1 = 0.0
        term2 = 0.0

        if denom1 != 0:
            term1 = ((t - knots[i]) / denom1) * self._basis_function(i, p - 1, t, knots)
        if denom2 != 0:
            term2 = ((knots[i+p+1] - t) / denom2) * self._basis_function(i+1, p - 1, t, knots)

        return term1 + term2


# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This file contains i/o headers for the SimVascular 1D solver (https://github.com/SimVascular/oneDSolver).
"""


class Headers(object):
    """
    Collection of headers of all oneDSolver input-file sections
    """
    node = [
        "#",
        "# ==========",
        "# NODE CARD",
        "# ==========",
        "# - Node Name (double)",
        "# - Node X Coordinate (double)",
        "# - Node Y Coordinate (double)",
        "# - Node Z Coordinate (double)",
        ""]

    joint1 = [
        "#",
        "# ==========",
        "# JOINT CARD",
        "# ==========",
        "# - Joint Name (string)",
        "# - Joint Node (double)",
        "# - Joint Inlet Name (string)",
        "# - Joint Outlet Name (string)",
        ""]

    joint2 = [
        "#",
        "# ================================",
        "# JOINTINLET AND JOINTOUTLET CARDS",
        "# ================================",
        "# - Inlet/Outlet Name (string)",
        "# - Total Number of segments (int)",
        "# - List of segments (list of int)",
        ""]

    segment = [
        "# ============",
        "# SEGMENT CARD",
        "# ============",
        "# - Segment Name (string)",
        "# - Segment ID (int)",
        "# - Segment Length (double)",
        "# - Total Finite Elements in Segment (int)",
        "# - Segment Inlet Node (int)",
        "# - Segment Outlet Node (int)",
        "# - Segment Inlet Area (double)",
        "# - Segment Outlet Area (double)",
        "# - Segment Inflow Value (double)",
        "# - Segment Material (string)",
        "# - Type of Loss (string - 'NONE','STENOSIS','BRANCH_THROUGH_DIVIDING','BRANCH_SIDE_DIVIDING','BRANCH_THROUGH_CONVERGING',",
        "#                          'BRANCH_SIDE_CONVERGING','BIFURCATION_BRANCH')",
        "# - Branch Angle (double)",
        "# - Upstream Segment ID (int)",
        "# - Branch Segment ID (int)",
        "# - Boundary Condition Type (string - 'NOBOUND','PRESSURE','AREA','FLOW','RESISTANCE','RESISTANCE_TIME','PRESSURE_WAVE',",
        "#                                     'WAVE','RCR','CORONARY','IMPEDANCE','PULMONARY')",
        "# - Data Table Name (string)",
        ""]

    solveroptions = [
        "# ==================",
        "# SOLVEROPTIONS CARD",
        "# ==================",
        "# - Solver Time Step (double), ",
        "# - Steps Between Saves (int), ",
        "# - Max Number of Steps (int)",
        "# - Number of quadrature points for finite elements (int), ",
        "# - Name of Datatable for inlet conditions (string)",
        "# - Type of boundary condition (string - 'NOBOUND','PRESSURE','AREA','FLOW','RESISTANCE','RESISTANCE_TIME','PRESSURE_WAVE',",
        "#                                        'WAVE','RCR','CORONARY','IMPEDANCE','PULMONARY')",
        "# - Convergence tolerance (double), ",
        "# - Formulation Type (int - 0 Advective, 1 Conservative), ",
        "# - Stabilization (int - 0 No stabilization, 1 With stabilization)",
        ""]

    material = [
        "# =============",
        "# MATERIAL CARD",
        "# =============",
        "# - Material Name (string)",
        "# - Material Type (string - 'LINEAR','OLUFSEN')",
        "# - Material Density (double)",
        "# - Material Viscosity (double)",
        "# - Material PRef (double)",
        "# - Material Exponent (double)",
        "# - Material Parameter 1 (double)",
        "# - Material Parameter 2 (double)",
        "# - Material Parameter 3 (double)",
        ""]

    output = [
        "# ============",
        "# OUTPUT CARD",
        "# ============",
        "#",
        "# 1. Output file format. The following output types are supported:",
        "#\t\tTEXT. The output of every segment is written in separate text files for the flow rate, pressure, area and Reynolds number. The rows contain output values at varying locations along the segment while columns contains results at various time instants.",
        "#\t\tVTK. The results for all time steps are plotted to a 3D-like model using the XML VTK file format.",
        "# 2. VTK export option. Two options are available for VTK file outputs:",
        "#\t\t0 - Multiple files (default). A separate file is written for each saved increment. A pvd file is also provided which contains the time information of the sequence. This is the best option to create animations.",
        "#\t\t1 - The results for all time steps are plotted to a single XML VTK file.",
        ""]

    junction = [
        "#",
        "# ==========",
        "# JUNCTION MODEL CARD",
        "# ==========",
        "# - joint_name",
        "# - junction_type, i.e. NORMAL_JUNCTION, etc.",
        ""]

    element = [
        "# =============",
        "# ELEMENT CARD",
        "# =============",
        "# - 1d_segment_name",
        "# - 1d_segment_number / 0d_element_number",
        "# - average_radius",
        "# - length",
        "# - blood_density",
        "# - blood_dynamic_viscosity",
        "# - Eh/r",
        "# - inlet_node_number",
        "# - outlet_node_number",
        "# - vessel_vector_x_coor",
        "# - vessel_vector_y_coor",
        "# - vessel_vector_z_coor",
        "# - boundary_condition_type",
        "# - boundary_condition_datatable_name",
        "# - 0d_element_type, i.e. R, RC, RL, RCL, L, C, etc.",
        "# - 0d_element_values in format of: R or R C or R L or R C L or L or C or blank for custom 0d elements",
        ""]

    inlet_bc = [
        "# ==============================",
        "# INLET BOUNDARY CONDITION CARD",
        "# ==============================",
        "# - inlet_segment_number",
        "# - inlet_boundary_condition_type",
        "# - inlet_boundary_condition_datatable_name",
        ""]

    solveroptions_0d = [
        "# ==================",
        "# SOLVEROPTIONS_0D CARD",
        "# ==================",
        "# - number_of_time_pts_per_cardiac_cycle (int)",
        "# - number_of_cardiac_cycles (int)",
        ""]

    def model(self, model_name):
        """
        Write model header
        """
        header = ["# ================================",
                  "# " + model_name + " MODEL - UNITS IN CGS",
                  "# ================================",
                  "",
                  "# ==========",
                  "# MODEL CARD",
                  "# ==========",
                  "# - Name of the model (string)",
                  "",
                  "MODEL " + model_name,
                  ""]
        return header
